!-----------------------------------------------------------------------
!
! Internal cluster routine to compute boxs as described by the
!  algorithm given in the Berger & Rigoutsos paper (IEEE Transactions on
!  systems, man, and cybernetics. VOL 21, NO. 5, Sept/Oct 1991).
!
!-----------------------------------------------------------------------
      integer function cmp_clstr(nx,ny,nz,box_max,blk_wd,box,box_stk,   &
     &                         sigx,sigy,sigz,lplx,lply,lplz,min_eff)

         implicit none

         integer  nx, ny, nz, box_max, blk_wd
         integer  box_stk(6,box_max)
         integer  sigx(nx),sigy(ny),sigz(nz),lplx(nx),lply(ny),lplz(nz)
         real*8   min_eff
         integer  box(nx,ny,nz)

! local variables
         integer  i, flag, no_box, j, ind, x_wd, y_wd, z_wd, k
         integer  l,m,n
         real*8   box_eff
         real*8   cur_box_eff
         character*20 fname


! initialize base box
         call InitStack(box_stk, 1, 1, 1, nx, ny, nz, no_box, box_max)

         i=1
 100     continue

! calculate efficiency of the box that will be examined
            cur_box_eff = box_eff(box, box_stk(1,i), nx, ny, nz)


!  If the efficiency of the current box is less than the threshold 
!   try to improve by splitting.
 
            if (cur_box_eff.lt.min_eff) then

! calculate signature and laplacian before split the box
               call cmp_sig_lpl(box,sigx,sigy,sigz,                &
     &                      lplx,lply,lplz,box_stk(1,i),nx,ny,nz)

               flag = 0
               call find_split(nx,ny,nz,no_box,i,blk_wd,        &
     &                        box_stk,                          &
     &                        sigx, sigy, sigz, lplx, lply,     &
     &                        lplz, flag, box_max)

! if flag is zero i.e. cannot split the current box, we examine the next one
               if (flag.eq.0) then
                   i = i + 1
               end if

            else
! if current box satisfies min. eff. given as input, we examine the next one
               i = i + 1
            end if


! # of box
            write(11,*) 'i= ',i,'    no_box= ', no_box

 
!  keep trying to divide the inefficient boxes until all boxes
!   satisfies minimum efficiency requirement
 
         if (i .le. no_box .and. no_box.lt.box_max) then
            goto 100
         end if

         cmp_clstr = no_box
 
      return
      end
!-----------------------------------------------------------------------
!
! subroutine that computes the signatures and laplacians of the given box\
!
!-----------------------------------------------------------------------
      subroutine cmp_sig_lpl(box, sigx, sigy, sigz, lplx, lply,     &
     &                       lplz, box_stk, nx, ny, nz)

      implicit none

         integer nx, ny, nz
         integer sigx(nx),sigy(ny),sigz(nz),box_stk(6)
         integer lplx(nx),lply(ny),lplz(nz)
         integer  box(nx,ny,nz)

         integer i, j, k,x_lo, y_lo, z_lo, x_hi, y_hi, z_hi, tmpl, tmpr


         call ivls(sigx,0,nx)
         call ivls(sigy,0,ny)
         call ivls(sigz,0,nz)
         x_lo = box_stk(1)
         y_lo = box_stk(2)
         z_lo = box_stk(3)
         x_hi = box_stk(4)
         y_hi = box_stk(5)
         z_hi = box_stk(6)

         do i = x_lo, x_hi
            do j = y_lo, y_hi
               do k = z_lo, z_hi
                  sigx(i) = sigx(i) + box(i,j,k)
               end do
            end do
         end do

         do j = y_lo, y_hi
            do i = x_lo, x_hi
               do k = z_lo, z_hi
                  sigy(j) = sigy(j) + box(i,j,k)
               end do
            end do
         end do

         do k = z_lo, z_hi
            do j = y_lo, y_hi
               do i = x_lo, x_hi
                  sigz(k) = sigz(k) + box(i,j,k)
               end do
            end do
         end do

! compute laplacians by the difference of 
! the absolute value of the gradients
!  - this should take care of some of 
! the problems of using the above laplacian.

         call ivls(lplx,0,nx)
         call ivls(lply,0,ny)
         call ivls(lplz,0,nz)

         do i = x_lo, x_hi-1
            do j = y_lo, y_hi
               do k = z_lo, z_hi
                  lplx(i) = lplx(i) + abs(box(i,j,k)-box(i+1,j,k))
               end do
            end do
         end do

         do j = y_lo, y_hi-1
            do i = x_lo, x_hi
               do k = z_lo, z_hi
                  lply(j) = lply(j) + abs(box(i,j,k)-box(i,j+1,k))
               end do
            end do
         end do

         do k = z_lo, z_hi-1
            do j = y_lo, y_hi
               do i = x_lo, x_hi
                  lplz(k) = lplz(k) + abs(box(i,j,k)-box(i,j,k+1))
               end do
            end do
         end do

         do i = x_hi-1,x_lo+1,-1
            lplx(i) = lplx(i)-lplx(i-1)
         end do

         do j = y_hi-1,y_lo+1,-1
            lply(j) = lply(j)-lply(j-1)
         end do

         do k = z_hi-1,z_lo+1,-1
            lplz(k) = lplz(k)-lplz(k-1)
         end do

         lplx(x_lo) = 0
         lply(y_lo) = 0
         lplz(z_lo) = 0
         lplx(x_hi) = 0
         lply(y_hi) = 0
         lplz(z_hi) = 0

      return
      end
!-----------------------------------------------------------------------
!
!  subroutine that computes the efficiency of the given box
!   box_eff = (number of flagged mesh points in the box)/
!              (total number of mesh points in the box)
!
!-----------------------------------------------------------------------
      double precision function box_eff(box, box_stk, nx, ny, nz)

      implicit none

         integer nx, ny, nz
         integer box_stk(6)
         integer box(nx,ny,nz)

         integer flg_pts, i, j, k, x_lo, y_lo, z_lo, x_hi, y_hi, z_hi

         x_lo = box_stk(1)
         y_lo = box_stk(2)
         z_lo = box_stk(3)
         x_hi = box_stk(4)
         y_hi = box_stk(5)
         z_hi = box_stk(6)

         flg_pts = 0

         do i = x_lo, x_hi
            do j = y_lo, y_hi
               do k = z_lo, z_hi
                  flg_pts = flg_pts + (box(i,j,k))
               end do
            end do
         end do

         box_eff = 1.0d0*flg_pts/(x_hi-x_lo+1.0d0)/                   &
     &              (y_hi-y_lo+1.0d0)/(z_hi-z_lo+1.0d0)

      return
      end

!-----------------------------------------------------------------------
!
!  This routine tries to split the inefficient box first by 
!   seeing if there are any zeros in the signatures (lines in the
!   box mesh with no flagged points) and if that does not
!   work then it computes the laplacian of the signatures and looks
!   for inflection points (i.e. where the laplacian goes to zero).
!
!-----------------------------------------------------------------------
      subroutine find_split(nx,ny,nz,no_box,stkid,blk_wd,               &
     &                     box_stk,                                     &
     &                     sigx, sigy, sigz, lplx, lply, lplz,          &
     &                     flag, box_max)

      implicit none

         integer   Find_Holes, Find_Inflx
         integer   no_box, nx, ny, nz, stkid, blk_wd
         integer   flag, box_max
         integer   sigx(nx), sigy(ny), sigz(nz)
         integer   lplx(nx), lply(ny), lplz(nz)
         integer   box_stk(6,box_max)

 
!       Look for Holes
!       Search for holes and split box and store in box_stk.
!       flag returns 1 if holes were found. 0 if not.
!       Routine has to add to NO_box if splitting is done.
 
         flag = find_holes(no_box,stkid,blk_wd,sigx,sigy,sigz,          &
     &                      box_stk,box_max)

         write(11,*) 'flag after Find_holes =',flag

 
!       If no holes are found then look for inflextion points.
!           Search for inflextion points and split if possible.
!         If splitting is possible add to stack. Return flag = 1
!         if splitting is carried out. else return 0
!       Routine has to add to NO_box if splitting is done.
 
         if ( flag .eq. 0 )then
            flag = Find_Inflx(nx,ny,nz,no_box,stkid,blk_wd,sigx,         &
     &                  sigy,sigz,lplx,lply,lplz,box_stk,box_max)
         end if

         write(11,*) 'flag after Find_Inflx =',flag

      return
      end

!-----------------------------------------------------------------------
!
! Integer function that attempts to find a hole (zero in the signature
!  array) in the box mesh and if it does then it splits the
!  box into two at that hole
!
!-----------------------------------------------------------------------
      integer function find_holes(no_box,stkid,blk_wd,                  &
     &                           sigx, sigy, sigz, box_stk,box_max)

      implicit none

         integer no_box, stkid, box_max, blk_wd
         integer sigx(*), sigy(*), sigz(*), box_stk(6,*)
         integer i, j, k, nx_lo, nx_hi, ny_lo, ny_hi, nz_lo, nz_hi


         find_holes = 0
         nx_lo = box_stk(1,stkid)
         ny_lo = box_stk(2,stkid)
         nz_lo = box_stk(3,stkid)
         nx_hi = box_stk(4,stkid)
         ny_hi = box_stk(5,stkid)
         nz_hi = box_stk(6,stkid)

! first, see if the box can be reduced
         i = nx_lo
 10      if (sigx(i).eq.0.and.(nx_hi-nx_lo).gt.blk_wd) then
               nx_lo = i + 1
            i = i + 1
            goto 10
         end if

         i = nx_hi
 15      if (sigx(i).eq.0.and.(nx_hi-nx_lo).gt.blk_wd) then
               nx_hi = i - 1
            i = i - 1
            goto 15
         end if

         j = ny_lo
  20     if (sigy(j).eq.0.and.(ny_hi-ny_lo).gt.blk_wd) then
               ny_lo = j + 1
            j = j + 1
            goto 20
         end if

         j = ny_hi
  25     if (sigy(j).eq.0.and.(ny_hi-ny_lo).gt.blk_wd) then
               ny_hi = j - 1
            j = j - 1
            goto 25
         end if

         k = nz_lo
  30     if (sigz(k).eq.0.and.(nz_hi-nz_lo).gt.blk_wd) then
               nz_lo = k + 1
            k = k + 1
            goto 30
         end if

         k = nz_hi
  35     if (sigz(k).eq.0.and.(nz_hi-nz_lo).gt.blk_wd) then
               nz_hi = k - 1
            k = k - 1
            goto 35
         end if

! After reducing
         box_stk(1,stkid) = nx_lo
         box_stk(2,stkid) = ny_lo
         box_stk(3,stkid) = nz_lo
         box_stk(4,stkid) = nx_hi
         box_stk(5,stkid) = ny_hi
         box_stk(6,stkid) = nz_hi


! now see if the current box can be divided

         do i = nx_lo+blk_wd+1,nx_hi-blk_wd-1
            if (sigx(i).eq.0) then


               call boxStack(stkid,nx_lo,ny_lo,nz_lo,i-1, ny_hi,    &
     &                        nz_hi,i+1,ny_lo,nz_lo,nx_hi,ny_hi,    &
     &                        nz_hi, box_stk, no_box, box_max)

               Find_Holes = 1
               goto 200
            end if
         end do

         do j = ny_lo+blk_wd+1,ny_hi-blk_wd-1
            if (sigy(j).eq.0) then

               call boxStack(stkid,nx_lo,ny_lo,nz_lo,nx_hi,j-1,      &
     &                        nz_hi,nx_lo,j+1,nz_lo,nx_hi,ny_hi,     &
     &                        nz_hi, box_stk, no_box, box_max)

               Find_Holes = 1
               goto 200
            end if
         end do

         do k = nz_lo+blk_wd+1,nz_hi-blk_wd-1
            if (sigz(k).eq.0) then

               call boxStack(stkid, nx_lo,ny_lo,nz_lo,nx_hi,ny_hi, &
     &                        k-1,nx_lo,ny_lo,k+1,nx_hi,ny_hi,     &
     &                        nz_hi, box_stk, no_box, box_max)

               Find_Holes = 1
               goto 200
            end if
         end do
 200  continue

      return
      end

          
!
!  Routine: boxStack 
!   Internal Routine to Add and Delete boxs from the 
!    box stack.
!   
!  Input:
!       stkid := Current stack number. The one to delete.
!       stlo1x/y/z := First box begining coordinates. 
!       stlo2x/y/z := Second box begining coordinates. 
!       sthi1x/y/z := First box extensions. 
!       sthi2x/y/z := Second box extensions. 
!       box_stk := box Stack.
!       NO_box  := Number of boxs thus far. Will add 1 to this.
!       box_max     := Maximum number of boxs allowed.
!  
!  Output:
!       Routine will update the value of NO_box and add in an
!       additional box and replace the current one in box_stk.
!
!-----------------------------------------------------------------------
      subroutine boxStack(stkid, stlo1x, stlo1y, stlo1z,sthi1x, sthi1y,   &
     &           sthi1z,stlo2x, stlo2y, stlo2z, sthi2x, sthi2y, sthi2z,   &
     &           box_stk, NO_box, box_max)

      implicit none

         integer NO_box, box_max
         integer stkid
         integer stlo1x, stlo1y, stlo1z, sthi1x, sthi1y, sthi1z
         integer stlo2x, stlo2y, stlo2z, sthi2x, sthi2y, sthi2z
         integer box_stk(6,box_max)
         integer i,j

! Check if adding one more box will exceed the maximum stack size. 
         if( (NO_box + 1) .gt. box_max )then
           write(6,*)'Error in boxStack:'
           write(6,*)'   Number of boxs will exceed maximum  stack size'
           stop
         end if

! Begin by replacing the current stack with the first box.
         box_stk( 1, stkid) = stlo1x
         box_stk( 2, stkid) = stlo1y
         box_stk( 3, stkid) = stlo1z
         box_stk( 4, stkid) = sthi1x
         box_stk( 5, stkid) = sthi1y
         box_stk( 6, stkid) = sthi1z

! Next add one more stack at the end.
         box_stk( 1, NO_box+1) = stlo2x
         box_stk( 2, NO_box+1) = stlo2y
         box_stk( 3, NO_box+1) = stlo2z
         box_stk( 4, NO_box+1) = sthi2x
         box_stk( 5, NO_box+1) = sthi2y
         box_stk( 6, NO_box+1) = sthi2z

!     Add one to the number of boxs
         NO_box = NO_box + 1


       return
       end

!-----------------------------------------------------------------------
! Routine: InitStack
!
! Simple Routine that initializes the box stack.
!
! Input: 
!      box_stk := box Stack array.
!       stbgx/y := box begining coordinates.
!       stexx/y := box extensions.
!-----------------------------------------------------------------------
      subroutine InitStack(box_stk, stbgx, stbgy, stbgz,stexx, stexy,    &
     &           stexz,NO_box, box_max)

         implicit none

         integer box_max, NO_box, box_stk(6,box_max),                    &
     &           stbgx, stbgy, stbgz,stexx, stexy,stexz


! Initialize NO_box
         NO_box = 1

! Set up first stack element.
         box_stk( 1, 1) = stbgx
         box_stk( 2, 1) = stbgy
         box_stk( 3, 1) = stbgz
         box_stk( 4, 1) = stexx
         box_stk( 5, 1) = stexy
         box_stk( 6, 1) = stexz

        return
        end

!-----------------------------------------------------------------------
!
!   Routine: Find_Inflx
!   Integer function that calculates the laplacian given the signature
!   arrays and locates inflextion points or zeros in the laplacian.
!   With these inflextion points a choice is made as to how to split the
!   box.
!
!   Laplacian array begins at lox+1 and ends at hix-1. 
!
!-----------------------------------------------------------------------
      integer function find_inflx(nx,ny,nz,no_box,stkid,blk_wd,sigx,    &
     &                      sigy,sigz,lplx,lply,lplz,box_stk, box_max)

      implicit none

         integer no_box,stkid,box_max,nx,ny,nz,blk_wd
         integer box_stk(6,box_max)
         integer sigx(nx), sigy(ny), sigz(nz)
         integer lplx(nx), lply(ny), lplz(nz)

         integer nx_lo, ny_lo, nz_lo, nx_hi, ny_hi, nz_hi
         integer nx_lo1, ny_lo1, nz_lo1,nx_hi1, ny_hi1,nz_hi1
         integer nx_lo2, ny_lo2, nz_lo2, nx_hi2, ny_hi2, nz_hi2

         integer i,j, k, hi_lplx, infptx, hi_lply, infpty, tmp
         integer hi_lplz, infptz

         nx_lo = box_stk(1,stkid)
         ny_lo = box_stk(2,stkid)
         nz_lo = box_stk(3,stkid)
         nx_hi = box_stk(4,stkid)
         ny_hi = box_stk(5,stkid)
         nz_hi = box_stk(6,stkid)

 
! looking for two possible cases:
!   two succesive laplacians have opposite signs
         hi_lplx = 0
         infptx = 0
         do i = nx_lo+blk_wd+1,nx_hi-blk_wd-1
            if ((lplx(i).gt.0.and.lplx(i+1).lt.0).or.           &
     &          (lplx(i).lt.0.and.lplx(i+1).gt.0)) then
                  tmp = abs(lplx(i)-lplx(i+1))
                  if (tmp.gt.hi_lplx) then
                     hi_lplx = tmp
                     infptx  = i
                  end if
            end if
         end do

         hi_lply = 0
         infpty = 0
         do j = ny_lo+blk_wd+1,ny_hi-blk_wd-1
            if ((lply(j).gt.0.and.lply(j+1).lt.0).or.            &
     &          (lply(j).lt.0.and.lply(j+1).gt.0)) then
                  tmp = abs(lply(j)-lply(j+1))
                  if (tmp.gt.hi_lply) then
                     hi_lply = tmp
                     infpty  = j
                  end if
            end if
         end do

         hi_lplz = 0
         infptz = 0
         do k = nz_lo+blk_wd+1,nz_hi-blk_wd-1
            if ((lplz(k).gt.0.and.lplz(k+1).lt.0).or.           &
     &          (lplz(k).lt.0.and.lplz(k+1).gt.0)) then
                  tmp = abs(lplz(k)-lplz(k+1))
                  if (tmp.gt.hi_lplz) then
                     hi_lplz = tmp
                     infptz  = k
                  end if
            end if
         end do

         if (infpty.ne.0.or.infptx.ne.0.or.infptz.ne.0) then
            if (hi_lplx.ge.hi_lply.and.hi_lplx.ge.hi_lplz) then
               nx_lo1 = nx_lo
               ny_lo1 = ny_lo
               nz_lo1 = nz_lo
               nx_hi1 = infptx
               ny_hi1 = ny_hi
               nz_hi1 = nz_hi

               nx_lo2 = infptx+1
               ny_lo2 = ny_lo
               nz_lo2 = nz_lo
               nx_hi2 = nx_hi
               ny_hi2 = ny_hi
               nz_hi2 = nz_hi

            else if (hi_lply.ge.hi_lplx.and.hi_lply.ge.hi_lplz) then
               nx_lo1 = nx_lo
               ny_lo1 = ny_lo
               nz_lo1 = nz_lo
               nx_hi1 = nx_hi
               ny_hi1 = infpty
               nz_hi1 = nz_hi

               nx_lo2 = nx_lo
               ny_lo2 = infpty+1
               nz_lo2 = nz_lo
               nx_hi2 = nx_hi
               ny_hi2 = ny_hi
               nz_hi2 = nz_hi

            else
               nx_lo1 = nx_lo
               ny_lo1 = ny_lo
               nz_lo1 = nz_lo
               nx_hi1 = nx_hi
               ny_hi1 = ny_hi
               nz_hi1 = infptz

               nx_lo2 = nx_lo
               ny_lo2 = ny_lo
               nz_lo2 = infptz+1
               nx_hi2 = nx_hi
               ny_hi2 = ny_hi
               nz_hi2 = nz_hi
            end if
               
            call boxStack(stkid, nx_lo1, ny_lo1, nz_lo1,nx_hi1, ny_hi1,  &
     &           nz_hi1, nx_lo2, ny_lo2, nz_lo2,nx_hi2, ny_hi2, nz_hi2,  &
     &           box_stk, NO_box, box_max)
            find_inflx = 1
         else
            find_inflx = 0
         end if

      return
      end
!---------------------------------------------------------------------
!
! Load vector with scalar.
!
!---------------------------------------------------------------------
      SUBROUTINE IVLS(V1,S1,N)

         INTEGER     V1(1),S1,I,N

         DO I = 1 , N
            V1(I) = S1
         enddo

      RETURN
      END
