/* binio.c */

/*
 Orignal authors of this file are
    Bill Hibbard, Brian Paul, Dave Santek, and Andre Battaiola.
*/

/*
 * The file format is assumed to be BIG-ENDIAN.
 * If this code is compiled with -DLITTLE and executes on a little endian
 * CPU then byte-swapping will be done.
 *
 * If an ANSI compiler is used prototypes and ANSI function declarations
 * are used.  Otherwise use K&R conventions.
 *
 * If we're running on a CRAY (8-byte ints and floats), conversions will
 * be done as needed.
 */

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#ifdef cray
#  include <string.h>
#endif
#include "binio.h"


/**********************************************************/
/*                  Byte Flipping                         */
/**********************************************************/


#define FLIP4( n )  (  (n & 0xff000000) >> 24		\
		     | (n & 0x00ff0000) >> 8		\
		     | (n & 0x0000ff00) << 8		\
		     | (n & 0x000000ff) << 24  )


#define FLIP2( n )  (((unsigned short) n & 0xff00) >> 8  |  (n & 0x00ff) << 8)


/*
 * Flip the order of the 4 bytes in an array of 4-byte words.
 */
#ifdef __STDC__
  void flip4( unsigned int *src, unsigned int *dest, int n )
#else
  void flip4( src, dest, n )
  unsigned int *src, *dest;
  int n;
#endif
{
   int i;
   for (i=0;i<n;i++) {
      register unsigned int tmp = src[i];
      dest[i] = FLIP4( tmp );
   }
}


/*
 * Flip the order of the 2 bytes in an array of 2-byte words.
 */
#ifdef __STDC__
  void flip2( unsigned short *src, unsigned short *dest, int n )
#else
  void flip2( src, dest, n )
  unsigned short *src, *dest;
  int n;
#endif
{
   int i;

   for (i=0;i<n;i++) {
      register unsigned short tmp = src[i];
      dest[i] = FLIP2( tmp );
   }
}


#ifdef cray

/** THESE ROUTINES MUST BE COMPILED ON THE CRAY ONLY SINCE THEY **/
/** REQUIRE 8-BYTES PER C-TYPE LONG                             **/

/* Cray to IEEE single precision */
static void c_to_if(t,f)
long *t,*f;
{
    if (*f != 0){
        *t = (((*f & 0x8000000000000000) |      /* sign bit */
                 ((((*f & 0x7fff000000000000) >> 48)-16258) << 55)) + /* exp */
                 (((*f & 0x00007fffff000000) +
                    ((*f & 0x0000000000800000) << 1)) << 8));  /* mantissa */
    }
    else *t = *f;
}


#define C_TO_IF( T, F )							\
	if (F != 0) {							\
		T = (((F & 0x8000000000000000) |			\
		((((F & 0x7fff000000000000) >> 48)-16258) << 55)) +	\
		(((F & 0x00007fffff000000) +				\
		((F & 0x0000000000800000) << 1)) << 8));		\
	}								\
	else {								\
		T = F;							\
	}



/* IEEE single precison to Cray */
static void if_to_c(t,f)
long *t,*f;
{
    if (*f != 0) {
        *t = (((*f & 0x8000000000000000) |
                ((*f & 0x7f80000000000000) >> 7) +
                (16258 << 48)) |
                (((*f & 0x007fffff00000000) >> 8) | (0x0000800000000000)));
        if ((*f << 1) == 0) *t = 0;
    }
    else *t = *f;
}

/* T and F must be longs! */
#define IF_TO_C( T, F )							\
	if (F != 0) {							\
		T = (((F & 0x8000000000000000) |			\
		((F & 0x7f80000000000000) >> 7) +			\
		(16258 << 48)) |					\
		(((F & 0x007fffff00000000) >> 8) | (0x0000800000000000)));  \
		if ((F << 1) == 0) T = 0;				\
	}								\
	else {								\
		T = F;							\
	}


/*
 * Convert an array of Cray 8-byte floats to an array of IEEE 4-byte floats.
 */
#ifdef __STDC__
void cray_to_ieee_array( long *dest, float *source, int n )
#else
void cray_to_ieee_array( dest, source, n )
long *dest;
float *source;
int n;
#endif
{
   long *dst, *src;
   long tmp1, tmp2;
   int i;

   dst = dest;
   src = (long *) source;

   for (i=0;i<n;i+=2) {       /* add 1 in case n is odd */
      c_to_if( &tmp1, &src[i] );
      c_to_if( &tmp2, &src[i+1] );
      *dst = (tmp1 & 0xffffffff00000000) | (tmp2 >> 32);
      dst++;
   }
}


/*
 * Convert an array of IEEE 4-byte floats to an array of 8-byte Cray floats.
 */
#ifdef __STDC__
void ieee_to_cray_array( float *dest, long *source, int n )
#else
void ieee_to_cray_array( dest, source, n )
float *dest;
long *source;
int n;
#endif
{
   long *dst, *src;
   int i;
   long ieee;

   src = source;
   dst = (long *) dest;

   for (i=0;i<n;i++) {
      /* most significant 4-bytes of ieee contain bit pattern to convert */
      if ((i&1)==0) {
	 /* get upper half */
	 ieee = src[i/2] & 0xffffffff00000000;
      }
      else {
	 /* get lower half */
	 ieee = src[i/2] << 32;
      }
      if_to_c( dst, &ieee );
      dst++;
   }
}


#endif /*cray*/



/**********************************************************************/
/*****                     Read Functions                         *****/
/**********************************************************************/


/*
 * Read a block of bytes.
 *  Input:  f - the file descriptor to read from.
 *         b - address of buffer to read into.
 *         n - number of bytes to read.
 * Return:  number of bytes read, 0 if error.
 */
#ifdef __STDC__
  int read_bytes( int f, void *b, int n )
#else
  int read_bytes( f, b, n )
  int f;
  void *b;
  int n;
#endif
{
   return read( f, b, n );
}



/*
 * Read an array of 2-byte integers.
 * Input:  f - file descriptor
 *         iarray - address to put integers
 *         n - number of integers to read.
 * Return:  number of integers read.
 */
#ifdef __STDC__
  int read_int2_array( int f, short *iarray, int n )
#else
  int read_int2_array( f, iarray, n )
  int f;
  short *iarray;
  int n;
#endif
{
#ifdef cray
   int i;
   signed char *buffer;
   int nread;
   buffer = (signed char *) malloc( n * 2 );
   if (!buffer)  return 0;
   nread = read( f, buffer, n*2 );
   if (nread<=0)  return 0;
   nread /= 2;
   for (i=0;i<nread;i++) {
      /* don't forget about sign extension! */
      iarray[i] = (buffer[i*2] * 256) | buffer[i*2+1];
   }
   free( buffer );
   return nread;
#else
   int nread = read( f, iarray, n*2 );
   if (nread<=0)
      return 0;
#ifdef LITTLE
   flip2( iarray, iarray, nread/2 );
#endif
   return nread/2;
#endif
}



/*
 * Read an array of unsigned 2-byte integers.
 * Input:  f - file descriptor
 *         iarray - address to put integers
 *         n - number of integers to read.
 * Return:  number of integers read.
 */
#ifdef __STDC__
  int read_uint2_array( int f, unsigned short *iarray, int n )
#else
  int read_uint2_array( f, iarray, n )
  int f;
  unsigned short *iarray;
  int n;
#endif
{
#ifdef cray
   int i;
   unsigned char *buffer;
   int nread;
   buffer = (unsigned char *) malloc( n * 2 );
   if (!buffer)  return 0;
   nread = read( f, buffer, n*2 );
   if (nread<=0)  return 0;
   nread /= 2;
   for (i=0;i<nread;i++) {
      iarray[i] = (buffer[i*2] << 8) | buffer[i*2+1];
   }
   free( buffer );
   return nread;
#else
   int nread = read( f, iarray, n*2 );
   if (nread<=0)
      return 0;
#ifdef LITTLE
   flip2( iarray, iarray, nread/2 );
#endif
   return nread/2;
#endif
}



/*
 * Read a 4-byte integer.
 * Input:  f - the file descriptor to read from
 *         i - pointer to integer to put result into.
 * Return:  1 = ok, 0 = error
 */
#ifdef __STDC__
  int read_int4( int f, int *i )
#else
  int read_int4( f, i )
  int f;
  int *i;
#endif
{
#ifdef LITTLE
   /* read big endian and convert to little endian */
   unsigned int n;
   if (read( f, &n, 4 )==4) {
      *i = FLIP4( n );
      return 1;
   }
   else {
      return 0;
   }
#else
   if (read( f, i, 4 )==4) {
#  ifdef cray
      *i = *i >> 32;
#  endif
      return 1;
   }
   else {
      return 0;
   }
#endif
}



/*
 * Read an array of 4-byte integers.
 * Input:  f - file descriptor
 *         iarray - address to put integers
 *         n - number of integers to read.
 * Return:  number of integers read.
 */
#ifdef __STDC__
  int read_int4_array( int f, int *iarray, int n )
#else
  int read_int4_array( f, iarray, n )
  int f;
  int *iarray;
  int n;
#endif
{
#ifdef cray
   int j, nread;
   int *buffer;

   buffer = (int *) malloc( (n+1)*4 );
   if (!buffer)
      return 0;
   nread = read( f, buffer, 4*n );
   if (nread<=0) {
      return 0;
   }
   nread /= 4;

   for (j=0;j<nread;j++) {
      if ((j&1)==0) {
	 iarray[j] = buffer[j/2] >> 32;
      }
      else {
	 iarray[j] = buffer[j/2] & 0xffffffff;
      }
   }
   free( buffer );
   return nread;
#else
   int nread = read( f, iarray, 4*n );
   if (nread<=0)
     return 0;
#  ifdef LITTLE
      flip4( iarray, iarray, nread/4 );
#  endif
   return nread/4;
#endif
}



/*
 * Read a 4-byte IEEE float.
 * Input:  f - the file descriptor to read from.
 *         x - pointer to float to put result into.
 * Return:  1 = ok, 0 = error
 */
#ifdef __STDC__
  int read_float4( int f, float *x )
#else
  int read_float4( f, x )
  int f;
  float *x;
#endif
{
#ifdef cray
   long buffer = 0;

   if ( read( f, &buffer, 4 )==4 ) {
      /* convert IEEE float (buffer) to Cray float (x) */
      if_to_c( x, &buffer );
      return 1;
    }
    return 0;
#else
#  ifdef LITTLE
      unsigned int n, *iptr;
      if (read( f, &n, 4 )==4) {
	 iptr = (unsigned int *) x;
	 *iptr = FLIP4( n );
	 return 1;
      }
      else {
	 return 0;
      }
#  else
      if (read( f, x, 4 )==4) {
	 return 1;
      }
      else {
	 return 0;
      }
#  endif
#endif
}



/*
 * Read an array of 4-byte IEEE floats.
 * Input:  f - file descriptor
 *         x - address to put floats
 *         n - number of floats to read.
 * Return:  number of floats read.
 */
#ifdef __STDC__
  int read_float4_array( int f, float *x, int n )
#else
  int read_float4_array( f, x, n )
  int f;
  float *x;
  int n;
#endif
{
#ifdef cray
   /* read IEEE floats into buffer, then convert to Cray format */
   long *buffer;
   int i, nread;

   buffer = (long *) malloc( (n+1) * 4 );
   if (!buffer) return 0;
   nread = read( f, buffer, n*4 );
   if (nread<=0)  return 0;
   nread /= 4;
   ieee_to_cray_array( x, buffer, nread );
   free( buffer );
   return nread;
#else
   int nread = read( f, x, 4*n );
   if (nread<=0)
      return 0;
#ifdef LITTLE
   flip4( (unsigned int *) x, (unsigned int *) x, nread/4 );
#endif
   return nread/4;
#endif
}




/*
 * Read a block of memory.
 * Input:  f - file descriptor
 *         data - address of first byte
 *         elements - number of elements to read
 *         elsize - size of each element to read (1, 2 or 4)
 * Return: number of elements written
 */
#ifdef __STDC__
int read_block( int f, void *data, int elements, int elsize )
#else
int read_block( f, data, elements, elsize )
int f;
void *data;
int elements;
int elsize;
#endif
{
   if (elsize==1) {
      return read( f, data, elements );
   }
   else if (elsize==2) {
#ifdef LITTLE
      int n;
      n = read( f, data, elements*2 ) / 2;
      if (n==elements) {
         flip2( (unsigned short *) data, (unsigned short *) data, elements );
      }
      return n;
#else
      return read( f, data, elements*2 ) / 2;
#endif
   }
   else if (elsize==4) {
#ifdef LITTLE
      int n;
      n = read( f, data, elements*4 ) / 4;
      if (n==elements) {
         flip4( (unsigned int *) data, (unsigned int *) data, elements );
      }
      return n;
#else
      return read( f, data, elements*4 ) / 4;
#endif
   }
   else {
      printf("Fatal error in write_block(): bad elsize (%d)\n", elsize );
      abort();
   }
}



/**********************************************************************/
/*****                         Write Functions                    *****/
/**********************************************************************/



/*
 * Write a block of bytes.
 * Input:  f - the file descriptor to write to.
 *         b - address of buffer to write.
 *         n - number of bytes to write.
 * Return:  number of bytes written, 0 if error.
 */
#ifdef __STDC__
  int write_bytes( int f, void *b, int n )
#else
  int write_bytes( f, b, n )
  int f;
  void *b;
  int n;
#endif
{
   return write( f, b, n );
}




/*
 * Write an array of 2-byte integers.
 * Input:  f - file descriptor
 *         iarray - address to put integers
 *         n - number of integers to write.
 * Return:  number of integers written
 */
#ifdef __STDC__
  int write_int2_array( int f, short *iarray, int n )
#else
  int write_int2_array( f, iarray, n )
  int f;
  short *iarray;
  int n;
#endif
{
#ifdef cray
   printf("write_int2_array not implemented!\n");
   exit(1);
#else
   int nwritten;
#ifdef LITTLE
   flip2( iarray, iarray, n );
#endif
   nwritten = write( f, iarray, 2*n );
#ifdef LITTLE
   flip2( iarray, iarray, n );
#endif
   if (nwritten<=0)
      return 0;
   return nwritten/2;
#endif
}



/*
 * Write an array of 2-byte unsigned integers.
 * Input:  f - file descriptor
 *         iarray - address to put integers
 *         n - number of integers to write.
 * Return:  number of integers written
 */
#ifdef __STDC__
  int write_uint2_array( int f, unsigned short *iarray, int n )
#else
  int write_uint2_array( f, iarray, n )
  int f;
  unsigned short *iarray;
  int n;
#endif
{
#ifdef cray
   int i, nwritten;
   unsigned char *buffer;
   buffer = (unsigned char *) malloc( 2*n );
   if (!buffer)  return 0;
   for (i=0;i<n;i++) {
      buffer[i*2] = (iarray[i] >> 8) & 0xff;
      buffer[i*2+1] = iarray[i] & 0xff;
   }
   nwritten = write( f, buffer, 2*n );
   free( buffer );
   if (nwritten<=0)
      return 0;
   else
      return nwritten/2;
#else
   int nwritten;
#ifdef LITTLE
   flip2( iarray, iarray, n );
#endif
   nwritten = write( f, iarray, 2*n );
#ifdef LITTLE
   flip2( iarray, iarray, n );
#endif
   if (nwritten<=0)
      return 0;
   else
      return nwritten/2;
#endif
}



/*
 * Write a 4-byte integer.
 *Input:  f - the file descriptor
 *         i - the integer
 * Return:  1 = ok, 0 = error
 */
#ifdef __STDC__
  int write_int4( int f, int i )
#else
  int write_int4( f, i )
  int f;
  int i;
#endif
{
#ifdef cray
   i = i << 32;
   return write( f, &i, 4 ) > 0;
#else
#  ifdef LITTLE
     i = FLIP4( i );
#  endif
   return write( f, &i, 4 ) > 0;
#endif
}



/*
 * Write an array of 4-byte integers.
 * Input:  f - the file descriptor
 *         i - the array of ints
 *	   n - the number of ints in array
 *  Return:  number of integers written.
 */
#ifdef __STDC__
  int write_int4_array( int f, int *i, int n )
#else
  int write_int4_array( f, i, n )
  int f;
  int *i;
  int n;
#endif
{
#ifdef cray
   int j, nwritten;
   char *buf, *b, *ptr;

   b = buf = (char *) malloc( n*4 + 8 );
   if (!b)
      return 0;
   ptr = (char *) i;
   for (j=0;j<n;j++) {
      ptr += 4;      /* skip upper 4 bytes */
      *b++ = *ptr++;
      *b++ = *ptr++;
      *b++ = *ptr++;
      *b++ = *ptr++;
   }
   nwritten = write( f, buf, 4*n );
   free( buf );
   if (nwritten<=0)
      return 0;
   else
      return nwritten / 4;
#else
#  ifdef LITTLE
      int nwritten;
      flip4( i, i, n );
      nwritten = write( f, i, 4*n );
      flip4( i, i, n );
      if (nwritten<=0)
	 return 0;
      else
	return nwritten / 4;
#  else
      return write( f, i, 4*n ) / 4;
#  endif
#endif
}



/*
 * Write a 4-byte IEEE float.
 * Input:  f - the file descriptor
 *         x - the float
 * Return:  1 = ok, 0 = error
 */
#ifdef __STDC__
  int write_float4( int f, float x )
#else
  int write_float4( f, x )
  int f;
  float x;
#endif
{
#ifdef cray
   char buffer[8];
   c_to_if( buffer, &x );
   return write( f, buffer, 4 ) > 0;
#else
#  ifdef LITTLE
      float y;
      unsigned int *iptr = (unsigned int *) &y, temp;
      y = (float) x;
      temp = FLIP4( *iptr );
      return write( f, &temp, 4 ) > 0;
#  else
      float y;
      y = (float) x;
      return write( f, &y, 4 ) > 0;
#  endif
#endif
}



/*
 * Write an array of 4-byte IEEE floating point numbers.
 * Input:  f - the file descriptor
 *         x - the array of floats
 *         n - number of floats in array
 * Return:  number of floats written.
 */
#ifdef __STDC__
  int write_float4_array( int f, float *x, int n )
#else
  int write_float4_array( f, x, n )
  int f;
  float *x;
  int n;
#endif
{
#ifdef cray
   /* convert cray floats to IEEE and put into buffer */
   int nwritten;
   long *buffer;
   buffer = (long *) malloc( n*4 + 8 );
   if (!buffer)
      return 0;
   cray_to_ieee_array( buffer, x, n );
   nwritten = write( f, buffer, 4*n );
   free( buffer );
   if (nwritten<=0)
      return 0;
   else
      return nwritten / 4;
#else
#  ifdef LITTLE
      int nwritten;
      flip4( (unsigned int *) x, (unsigned int *) x, n );
      nwritten = write( f, x, 4*n );
      flip4( (unsigned int *) x, (unsigned int *) x, n );
      if (nwritten<=0)
	 return 0;
      else 
	 return nwritten / 4;
#  else
      return write( f, x, 4*n ) / 4;
#  endif
#endif
}



/*
 * Write a block of memory.
 * Input:  f - file descriptor
 *         data - address of first byte
 *         elements - number of elements to write
 *         elsize - size of each element to write (1, 2 or 4)
 * Return: number of elements written
 */
#ifdef __STDC__
int write_block( int f, void *data, int elements, int elsize )
#else
int write_block( f, data, elements, elsize )
int f;
void *data;
int elements;
int elsize;
#endif
{
   if (elsize==1) {
      return write( f, data, elements );
   }
   else if (elsize==2) {
#ifdef LITTLE
      int n;
      flip2( (unsigned short *) data, (unsigned short *) data, elements );
      n = write( f, data, elements*2 ) / 2;
      flip2( (unsigned short *) data, (unsigned short *) data, elements );
      return n;
#else
      return write( f, data, elements*2 ) / 2;
#endif
   }
   else if (elsize==4) {
#ifdef LITTLE
      int n;
      flip4( (unsigned int *) data, (unsigned int *) data, elements );
      n = write( f, data, elements*4 ) / 4;
      flip4( (unsigned int *) data, (unsigned int *) data, elements );
      return n;
#else
      return write( f, data, elements*4 ) / 4;
#endif
   }
   else {
      printf("Fatal error in write_block(): bad elsize (%d)\n", elsize );
      abort();
   }
}
/* sockio.c */

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h> 

#include "sockio.h"

/******************************************************************
 * initialize the socket
 * return socket descriptor. If failed return -1
 ******************************************************************/
int initialize()
{
   int sockfd;

   char *hostname;
   char *rawport;
   int port;

   struct  hostent *h;
   struct sockaddr_in serv_addr;

   /** 
    * first read system environment JSERHOST and JSERPORT
    */
   hostname = getenv("JSERHOST");
   rawport = getenv("JSERPORT");

   if (hostname == NULL) {
      printf("Unknown server host name error ...\n");
      return -1;
   }
   
   if (rawport == NULL) 
      port = 7654;
   else
      port = atoi(rawport);

   if ((h=gethostbyname(hostname)) == NULL) {  /* get the host info */
     /*      herror("gethostbyname"); */
      printf("Error gethostbyname ...\n");
      return -1;
   }

   /*
    * fill in the structure "serv_addr" with the address 
    * of the server that we want to connect with 
    */
   bzero((char *) &serv_addr, sizeof(serv_addr));
   serv_addr.sin_family = AF_INET;
   serv_addr.sin_addr.s_addr = 
                    inet_addr(inet_ntoa(*((struct in_addr *)h->h_addr)));
   serv_addr.sin_port = htons(port);

   /*
   printf("Host name  : %s\n", h->h_name);
   printf("IP Address : %s\n",inet_ntoa(*((struct in_addr *)h->h_addr)));
   */

   /*
    * open a TCP socket (an Internet stream socket).
    */
   if ( (sockfd = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
      printf("Can't open stream socket.\n");
      return -1;
   }

   /*
    * connect to server
    */
   if (connect(sockfd, (struct sockaddr *) &serv_addr, 
                                    sizeof(serv_addr)) < 0) {
      printf("Can't connect to server.\n");
      return -1;
   }

   return sockfd;

}





/* java_ser1d.c */

#include <stdio.h>
#include <stdlib.h>
#include "sockio.h"
#include "binio.h"

/*********************************************************
 * java_ser: connect to server and send the data
 *     
 *
 * return value (all functions)
 *   0 if success
 *   1 otherwise
 *
 * It will read server host name and port from JSERHOST and
 * JSERPORT environment.
 *********************************************************/



/*************************************************************
 * java_fser_
 * c program to link with fortran 90
 * This program writes the x and fx values to file.
 *
 * input parameters
 *   name - name of the file
 *   time - time value
 *   x    - x value array
 *   fx   - fx value array
 *   nx   - size of x and fx array
 ************************************************************/
/* ONLY SUPPORT BIG ENDIAN */
int java_fser_(name, time, x, fx, nx) 
  char *name;
  double *time;
  double x[];
  double fx[];
  int *nx;
{
   int size, i;
   FILE *fopen(), *fp;

   if ((fp = fopen(name, "a+b")) == NULL) {
      printf("Error: can't open output file ...\n");
      return 1;
   }

   /* first write time and nx */
   fwrite(time, sizeof(double), 1, fp);
   fwrite(nx, sizeof(int), 1, fp);

   size = *nx;
   
   /* write x and fx value */
   for (i=0; i<size; i++) {
      fwrite(&x[i], sizeof(double), 1, fp);
      fwrite(&fx[i], sizeof(double), 1, fp);
   }
   
   fclose(fp);
   return 0;
}


/*************************************************************
 * java_ser_
 * c program to link with fortran 90
 * This program writes the x and fx values to socket.
 *
 * input parameters
 *   name - name of the file, it will be used fore title
 *   time - time value
 *   x    - x value array
 *   fx   - fx value array
 *   nx   - size of x and fx array
 ************************************************************/
int java_ser_(name, time, x, fx, nx) 
  char *name;
  double *time;
  double x[];
  double fx[];
  int *nx;
{
   int sockfd;
   int size = *nx, name_length, i;
   double ft = *time;
   float ftime = (float)ft;
   int command = 10;

   sockfd = initialize();
   if (sockfd == -1)
     exit(1);

   /* first send command */
   write_int4(sockfd, command);

   
   /* send title information */
   name_length = strlen(name);
   write_int4(sockfd, name_length);
   write(sockfd, name, name_length);

   /* write time and nx */
   write_float4(sockfd, ftime);
   write_int4(sockfd, size);

   /* write x and fx value */
   for (i=0; i<size; i++) {
      write_float4(sockfd, (float)x[i]);
      write_float4(sockfd, (float)fx[i]);
   }
   
   close(sockfd);
   return 0;
}


/*************************************************************
 * java_mser_
 * c program to link with fortran 90
 * This program writes whole set of time data with the x and fx values
 * to socket.
 *
 * input parameters
 *   name - name of the file, it will be used for title
 *   nt   - number of time
 *   time - time value array
 *   x    - x value array
 *   fx   - fx value array
 *   nx   - size of x and fx array
 ************************************************************/
int java_mser_(name, nt, time, x, fx, nx) 
  char *name;
  int *nt;
  double time[];
  double x[];
  double fx[];
  int nx[];
{
   int sockfd;
   int ntsize = *nt;
   int name_length, i, j, index;
   int command = 12;

   sockfd = initialize();
   if (sockfd == -1)
     exit(1);

   /* first send command */
   write_int4(sockfd, command);
   
   /* send title information */
   name_length = strlen(name);
   write_int4(sockfd, name_length);
   write(sockfd, name, name_length);

   /* send nt */
   write_int4(sockfd, ntsize);

   /* write time data */
   index = 0;

   for (i=0; i<ntsize; i++) {
      write_float4(sockfd, (float)time[i]);

      write_int4(sockfd, nx[i]);
      for (j=0; j<nx[i]; j++) {
         write_float4(sockfd, (float)x[index]);
         write_float4(sockfd, (float)fx[index]);
         index++;
      }
   }
   
   close(sockfd);
   return 0;
}




/*************************************************************
 * java_bbser_
 * c program to link with fortran 90
 * This program writes the x, y and fxy values to socket.
 *
 * input parameters
 *   name - name of the file, it will be used fore title
 *   time - time value
 *   bb    - x min and max  value
 *   xmax    - x max value
 *   nx      - number of x points 
 *   fx     - fxy value array
 ************************************************************/
int java_bbser_(name, time, bb, nx, fx) 
  char *name;
  double *time;
  double bb[];
  int *nx;
  double fx[];
{
   int sockfd;
   int size = *nx, name_length, i;
   double ft = *time;
   float ftime = (float)ft;
   int command = 28;

   sockfd = initialize();
   if (sockfd == -1)
     exit(1);

   /* first send command */
   write_int4(sockfd, command);
   
   /* send title information */
   name_length = strlen(name);
   write_int4(sockfd, name_length);
   write(sockfd, name, name_length);

   /* write time and nx */
   write_float4(sockfd, ftime);
   write_int4(sockfd, size);
   write_float4(sockfd, (float)bb[0]);
   write_float4(sockfd, (float)bb[1]);



   /* write fx value */
   for (i=0; i<size; i++) {
      write_float4(sockfd, (float)fx[i]);
   }
   
   close(sockfd);
   return 0;
}


/*************************************************************
 * java_bbmser_
 * c program to link with fortran 90
 * This program writes whole set of time data with the x, y and fxy values
 * to socket.
 *
 * input parameters
 *   name - name of the file, it will be used fore title
 *   nt   - number of time
 *   time - time value array
 *   bb    - x min and max array
 *   nx    -   nx value array
 *   fx   - fxy value array
 ************************************************************/
int java_bbmser_(name, nt, time, bb, nx, fx) 
  char *name;
  int *nt;
  double time[];
  double bb[][4];
  int nx[];
  double fx[];

{
   int sockfd;
   int ntsize = *nt, name_length, i, j, index;
   int command = 29;

   sockfd = initialize();
   if (sockfd == -1)
     exit(1);

   /* first send command */
   write_int4(sockfd, command);
   
   /* send title information */
   name_length = strlen(name);
   write_int4(sockfd, name_length);
   write(sockfd, name, name_length);

   /* send nt */
   write_int4(sockfd, ntsize);


   index = 0;
   for (i=0; i<ntsize; i++) {
      write_float4(sockfd, (float)time[i]);

      write_int4(sockfd, nx[i]);
      write_float4(sockfd, (float)bb[i][0]);
      write_float4(sockfd, (float)bb[i][1]);
     
      for (j=0; j<nx[i]; j++) {
         write_float4(sockfd, (float)fx[index]);
         index++;
      }
   }
   
   close(sockfd);
   return 0;
}


/*************************************************************
 * java_pser_
 * c program to link with fortran 90
 * to be used for parallel machine
 * This program writes the x and fx values to socket.
 *
 * input parameters
 *   name - name of the file, it will be used for title
 *   time - time value
 *   x    - x value array
 *   fx   - fx value array
 *   nx   - size of x and fx array
 *   p    - total number of processors
 ************************************************************/
int java_pser_(name, time, x, fx, nx, p) 
  char *name;
  double *time;
  double x[];
  double fx[];
  int *nx;
  int *p;
{
   int sockfd;
   int size = *nx, name_length, i, pp = *p;
   double ft = *time;
   float ftime = (float)ft;
   int command = 17;

   sockfd = initialize();
   if (sockfd == -1)
     exit(1);

   /* first send command */
   write_int4(sockfd, command);

   
   /* send title information */
   name_length = strlen(name);
   write_int4(sockfd, name_length);
   write(sockfd, name, name_length);

   /* write time, nx, and p */
   write_float4(sockfd, ftime);
   write_int4(sockfd, size);
   write_int4(sockfd, pp);

   /* write x and fx value */
   for (i=0; i<size; i++) {
      write_float4(sockfd, (float)x[i]);
      write_float4(sockfd, (float)fx[i]);
   }
   
   close(sockfd);
   return 0;
}



int java_ser(char *name, double time, double x[], double y[], int n) {
   return java_ser_(name, &time, x, y, &n);
}

int java_mser(char *name, int nt, double time[], double x[], double fx[], int nx[]) {
   return java_mser_(name, &nt, time, x, fx, nx);
}

int java_bbser(char *name, double time, double bb[], int nx, double fx[]) {
  java_bbser_(name, &time, bb, &nx, fx);
}

int java_bbmser(char *name, int nt, double time[], double bb[][4], int nx[], double fx[]) {
   java_bbmser_(name, &nt, time, bb, nx, fx);
}

int java_pser(char *name, double time, double x[], double fx[], int nx, int p) {
   java_pser_(name, &time, x, fx, &nx, &p);
}
/* java_ser2d.c */

#include <stdio.h>
#include <stdlib.h>
#include "sockio.h"
#include "binio.h"

/*********************************************************
 * java_ser2d: connect to server and send the data
 *     
 *
 * return value (all functions)
 *   0 if success
 *   1 otherwise
 *
 * It will read server host name and port from JSERHOST and
 * JSERPORT environment.
 *********************************************************/


/*************************************************************
 * java_fser2d_
 * c program to link with fortran 90
 * This program writes the x, y and fxy values to file.
 *
 * input parameters
 *   name - name of the file
 *   time - time value
 *   x    - x value array
 *   y    - y value array
 *   fxy  - fxy value array
 *   nx   - size of x, y and fx array
 ************************************************************/
/* ONLY SUPPORT BIG ENDIAN */
int java_fser2d_(name, time, x, y, fxy, nx) 
  char *name;
  double *time;
  double x[];
  double y[];
  double fxy[];
  int *nx;
{
   int size, i;
   FILE *fopen(), *fp;

   if ((fp = fopen(name, "a+b")) == NULL) {
      printf("Error: can't open output file ...\n");
      return 1;
   }
   size = *nx;

   /* first write time and nx */
   fwrite(time, sizeof(double), 1, fp);
   fwrite(nx, sizeof(int), 1, fp);

   /* write x and fx value */
   for (i=0; i<size; i++) {
      fwrite(&x[i], sizeof(double), 1, fp);
      fwrite(&y[i], sizeof(double), 1, fp);
      fwrite(&fxy[i], sizeof(double), 1, fp);
   }
   
   fclose(fp);
   return 0;
}


/*************************************************************
 * java_ser2d_
 * c program to link with fortran 90
 * This program writes the x, y and fxy values to socket.
 *
 * input parameters
 *   name - name of the file, it will be used fore title
 *   time - time value
 *   x    - x value array
 *   y    - y value array
 *   fxy  - fxy value array
 *   nx   - size of x and fx array
 ************************************************************/

int java_ser2d_(name, time, x, y, fxy, nx) 
  char *name;
  double *time;
  double x[];
  double y[];
  double fxy[];
  int *nx;
{
   int sockfd;
   int size = *nx, name_length, i;
   double ft = *time;
   float ftime = (float)ft;
   int command = 15;

   sockfd = initialize();
   if (sockfd == -1)
     exit(1);

   /* first send command */
   write_int4(sockfd, command);

   /* send title information */
   name_length = strlen(name);
   write_int4(sockfd, name_length);
   write(sockfd, name, name_length);

   /* write time and nx */
   write_float4(sockfd, ftime);
   write_int4(sockfd, size);

   /* write x and fx value */
   for (i=0; i<size; i++) {
      write_float4(sockfd, (float)x[i]);
      write_float4(sockfd, (float)y[i]);
      write_float4(sockfd, (float)fxy[i]);
   }
   
   close(sockfd);
   return 0;
}


/*************************************************************
 * java_mser2d_
 * c program to link with fortran 90
 * This program writes whole set of time data with the x, y and fxy values
 * to socket.
 *
 * input parameters
 *   name - name of the file, it will be used fore title
 *   nt   - number of time
 *   time - time value array
 *   x    - x value array
 *   y    - y value array
 *   fxy   - fxy value array
 *   nx   - size of x and fx array
 ************************************************************/
int java_mser2d_(name, nt, time, x, y, fxy, nx) 
  char *name;
  int *nt;
  double time[];
  double x[];
  double y[];
  double fxy[];
  int nx[];
{

   int sockfd;
   int ntsize = *nt;
   int name_length, i, j, index;
   int command = 16;

   sockfd = initialize();
   if (sockfd == -1)
     exit(1);

   /* first send command */
   write_int4(sockfd, command);
   
   /* send title information */
   name_length = strlen(name);
   write_int4(sockfd, name_length);
   write(sockfd, name, name_length);

   /* send nt */
   write_int4(sockfd, ntsize);

   /* write time data */
   index = 0;

   for (i=0; i<ntsize; i++) {
      write_float4(sockfd, (float)time[i]);

      write_int4(sockfd, nx[i]);
      for (j=0; j<nx[i]; j++) {
         write_float4(sockfd, (float)x[index]);
         write_float4(sockfd, (float)y[index]);
         write_float4(sockfd, (float)fxy[index]);
         index++;
      }
   }
   
   close(sockfd);
   return 0;
}


/*************************************************************
 * java_bser2d_
 * c program to link with fortran 90
 * This program writes the x, y and fxy values to socket.
 * (reduced form of java_ser2d)
 *
 * input parameters
 *   name - name of the file, it will be used fore title
 *   time - time value
 *   x    - x value array
 *   nx   - x array size
 *   y    - y value array
 *   ny   - y array size
 *   fxy  - fxy value array
 ************************************************************/
int java_bser2d_(name, time, x, nx, y, ny, fxy) 
  char *name;
  double *time;
  double x[];
  int *nx;
  double y[];
  int *ny;
  double fxy[];
{
   int sockfd;
   int size, xsize = *nx, ysize = *ny, name_length, i;
   double ft = *time;
   float ftime = (float)ft;
   int command = 30;

   sockfd = initialize();
   if (sockfd == -1)
     exit(1);

   /* first send command */
   write_int4(sockfd, command);

   
   /* send title information */
   name_length = strlen(name);
   write_int4(sockfd, name_length);
   write(sockfd, name, name_length);

   /* write time */
   write_float4(sockfd, ftime);

   /* send x array */
   write_int4(sockfd, xsize);
   for (i=0; i<xsize; i++) 
      write_float4(sockfd, (float)x[i]);

   /* send y array */
   write_int4(sockfd, ysize);
   for (i=0; i<ysize; i++) 
      write_float4(sockfd, (float)y[i]);

   /* send fxy array */
   size = xsize * ysize;
   write_int4(sockfd, size);
   for (i=0; i<size; i++) 
      write_float4(sockfd, (float)fxy[i]);

   close(sockfd);
   return 0;
}


/*************************************************************
 * java_bmser2d_
 * c program to link with fortran 90
 * This program writes whole set of time data with the x, y and fxy values
 * to socket.
 *
 * input parameters
 *   name - name of the file, it will be used fore title
 *   nt   - number of time
 *   time - time value array
 *   x    - x value array
 *   nx   - size of x array
 *   y    - y value array
 *   ny   - size of y array 
 *   fxy   - fxy value array
 ************************************************************/
int java_bmser2d_(name, nt, time, x, nx, y, ny, fxy) 
  char *name;
  int *nt;
  double time[];
  double x[];
  int nx[];
  double y[];
  int ny[];
  double fxy[];
{
  
   int sockfd;
   int ntsize = *nt, size, name_length, i, j, index, xindex, yindex;
   int command = 31;

   sockfd = initialize();
   if (sockfd == -1)
     exit(1);

   /* first send command */
   write_int4(sockfd, command);

   /* send title information */
   name_length = strlen(name);
   write_int4(sockfd, name_length);
   write(sockfd, name, name_length);

   /* send nt */
   write_int4(sockfd, ntsize);

   /* write time data */
   index = 0;
   xindex = 0;
   yindex = 0;

   for (i=0; i<ntsize; i++) {
      write_float4(sockfd, (float)time[i]);

      write_int4(sockfd, nx[i]);
      for (j=0; j<nx[i]; j++) {
         write_float4(sockfd, (float)x[xindex]);
         xindex++;
      }

      write_int4(sockfd, ny[i]);
      for (j=0; j<ny[i]; j++) {
         write_float4(sockfd, (float)y[yindex]);
         yindex++;
      }

      size = nx[i] * ny[i];
      for (j=0; j<size; j++) {
         write_float4(sockfd, (float)fxy[index]);
         index++;
      }
   }
   
   close(sockfd);
   return 0;
}

/*************************************************************
 * java_bbser2d_
 * c program to link with fortran 90
 * sends 2d bounding box and fxy values to socket.
 *
 * input parameters
 *   name - name of the file, it will be used fore title
 *   time - time value
 *   bb    - x min, max, y min, and y max
 *   xmax    - x max value
 *   nx      - number of x points 
 *   ymin    - y min value
 *   ymax    - y max value
 *   ny      - number of y points 
 *   fxy     - fxy value array
 ************************************************************/
int java_bbser2d_(name, time, bb, nx, ny, fxy) 
  char *name;
  double *time;
  double bb[];
  int *nx;
  int *ny;
  double fxy[];
{
   int sockfd;
   int size, xsize = *nx, ysize = *ny, name_length, i;
   double ft = *time;
   float ftime = (float)ft;
   int command = 25;

   sockfd = initialize();
   if (sockfd == -1)
     exit(1);

   /* first send command */
   write_int4(sockfd, command);
   
   /* send title information */
   name_length = strlen(name);
   write_int4(sockfd, name_length);
   write(sockfd, name, name_length);

   size = xsize * ysize;
   /* write time, x and y size, and bb */
   write_float4(sockfd, ftime);
   write_int4(sockfd, xsize);
   write_int4(sockfd, ysize);
   write_float4(sockfd, (float)bb[0]);
   write_float4(sockfd, (float)bb[1]);
   write_float4(sockfd, (float)bb[2]);
   write_float4(sockfd, (float)bb[3]);

   /* write fx value */
   for (i=0; i<size; i++) {
      write_float4(sockfd, (float)fxy[i]);
   }
   
   close(sockfd);
   return 0;
}

/*************************************************************
 * java_bbmser2d_
 * c program to link with fortran 90
 * This program writes whole set of time data with the x, y and fxy values
 * to socket.
 *
 * input parameters
 *   name - name of the file, it will be used fore title
 *   nt   - number of time
 *   time - time value array
 *   bb    - x min, max, y min, and y max value array
 *   nx    -   nx value array
 *   ny    -   ny value array
 *   fxy   - fxy value array
 ************************************************************/
int java_bbmser2d_(name, nt, time, bb, nx, ny, fxy) 
  char *name;
  int *nt;
  double time[];
  double bb[][4];
  int nx[];
  int ny[];
  double fxy[];

{
   int sockfd;
   int ntsize = *nt, size, name_length, i, j, index;
   int command = 26;

   sockfd = initialize();
   if (sockfd == -1)
     exit(1);

   /* first send command */
   write_int4(sockfd, command);
   
   /* send title information */
   name_length = strlen(name);
   write_int4(sockfd, name_length);
   write(sockfd, name, name_length);

   /* send nt */
   write_int4(sockfd, ntsize);

   index = 0;
   for (i=0; i<ntsize; i++) {
      write_float4(sockfd, time[i]);
      write_int4(sockfd, nx[i]);
      write_int4(sockfd, ny[i]);
      write_float4(sockfd, (float)bb[i][0]);
      write_float4(sockfd, (float)bb[i][1]);
      write_float4(sockfd, (float)bb[i][2]);
      write_float4(sockfd, (float)bb[i][3]);
     
      size = nx[i] * ny[i];
      for (j=0; j<size; j++) {
         write_float4(sockfd, (float)fxy[index]);
         index++;
      }
   }
   
   close(sockfd);
   return 0;
}


/*************************************************************
 * java_pser_
 * c program to link with fortran 90
 * to be used for parallel machine
 * This program writes the x and fx values to socket.
 *
 * input parameters
 *   name - name of the file, it will be used for title
 *   time - time value
 *   x    - x value array
 *   fx   - fx value array
 *   nx   - size of x and fx array
 *   p    - total number of processors
 ************************************************************/
int java_pser2d_(name, time, x, y, fxy, nx, p) 
  char *name;
  double *time;
  double x[];
  double y[];
  double fxy[];
  int *nx;
  int *p;
{
   int sockfd;
   int size = *nx, name_length, i, pp = *p;
   double ft = *time;
   float ftime = (float)ft;
   int command = 18;

   sockfd = initialize();
   if (sockfd == -1)
     exit(1);

   /* first send command */
   write_int4(sockfd, command);
   
   /* send title information */
   name_length = strlen(name);
   write_int4(sockfd, name_length);
   write(sockfd, name, name_length);

   /* write time, nx, and p */
   write_float4(sockfd, ftime);
   write_int4(sockfd, size);
   write_int4(sockfd, pp);

   /* write x and fx value */
   for (i=0; i<size; i++) {
      write_float4(sockfd, (float)x[i]);
      write_float4(sockfd, (float)y[i]);
      write_float4(sockfd, (float)fxy[i]);
   }
   
   close(sockfd);
   return 0;
}



/*************************************************************
 * java_pbbser2d_
 * sends 2d bounding box and fxy values to socket.
 * used by parallel machine
 *
 * input parameters
 *   name - name of the file, it will be used fore title
 *   time - time value
 *   bb    - x min, max, y min, and y max
 *   xmax    - x max value
 *   nx      - number of x points 
 *   ymin    - y min value
 *   ymax    - y max value
 *   ny      - number of y points 
 *   fxy     - fxy value array
 *   p       - the number of total processors
 ************************************************************/
int java_pbbser2d_(name, time, bb, nx, ny, fxy, p) 
  char *name;
  double *time;
  double bb[];
  int *nx;
  int *ny;
  double fxy[];
  int *p;
{
   int sockfd;
   int size, xsize = *nx, ysize = *ny, name_length, i;
   double ft = *time;
   float ftime = (float)ft;
   int pp = *p;
   int command = 32;

   sockfd = initialize();
   if (sockfd == -1)
     exit(1);

   /* first send command */
   write_int4(sockfd, command);
   
   /* send title information */
   name_length = strlen(name);
   write_int4(sockfd, name_length);
   write(sockfd, name, name_length);

   size = xsize * ysize;
   /* write time, x and y size, and bb */
   write_float4(sockfd, ftime);
   write_int4(sockfd, xsize);
   write_int4(sockfd, ysize);
   write_float4(sockfd, (float)bb[0]);
   write_float4(sockfd, (float)bb[1]);
   write_float4(sockfd, (float)bb[2]);
   write_float4(sockfd, (float)bb[3]);
   write_int4(sockfd, pp);

   /* write fx value */
   for (i=0; i<size; i++) {
      write_float4(sockfd, (float)fxy[i]);
   }
   
   close(sockfd);
   return 0;
}



int java_ser2d(char *name, double time, double x[], double y[], double fxy[], int n) {
   return java_ser2d_(name, &time, x, y, fxy, &n);
}

int java_mser2d(char *name, int nt, double time[], double x[], double y[], double fxy[], int nx[]) {
   return java_mser2d_(name, &nt, time, x, y, fxy, nx);
}

int java_bser2d(char *name, double time, double x[], int nx, double y[], int ny, double fxy[]) {
  java_bser2d_(name, &time, x, &nx, y, &ny, fxy);
}

int java_bmser2d(char *name, int nt, double time[], double x[], int nx[], double y[], int ny[], double fxy[]) {
  java_bmser2d_(name, &nt, time, x, nx, y, ny, fxy);
}

int java_bbser2d(char *name, double time, double bb[], int nx, int ny, double fxy[]) {
  java_bbser2d_(name, &time, bb, &nx, &ny, fxy);
}

int java_bbmser2d(char *name, int nt, double time[], double bb[][4], int nx[], int ny[], double fxy[]) {
   java_bbmser2d_(name, &nt, time, bb, nx, ny, fxy);
}

int java_pser2d(char *name, double time, double x[], double y[], double fx[], int nx, int p) {
   java_pser2d_(name, &time, x, y, fx, &nx, &p);
}


int java_pbbser2d(char *name, double time, double bb[], int nx, int ny, double fxy[], int p) {
   java_pbbser2d_(name, &time, bb, &nx, &ny, fxy, &p);
}



