/*
 * implementation of sv2 api
 *
 * WARNING: Putting all code segmentations in a file is intentional.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h> 

#include "sv2_api.h"


#define RAW_DATA_FEED  201

#define POINT         101
#define VECTOR        102
#define CONNECTIVITY  103
#define BAR           104
#define TREE          105
#define TWOD          201
#define THREED        301

#define FLIP4(n)  (  (n & 0xff000000) >> 24		\
		   | (n & 0x00ff0000) >> 8		\
		   | (n & 0x0000ff00) << 8		\
		   | (n & 0x000000ff) << 24  )


/*
 * The file format is assumed to be BIG-ENDIAN.
 * If this code is compiled with -DLITTLE and executes on a little endian
 * CPU then byte-swapping will be done.
 *
 * If we're running on a CRAY (8-byte ints and floats), conversions will
 * be done as needed. (NEED SOME TESTS ....)
 */

/*
 * Flip the order of the 4 bytes in an array of 4-byte words.
 */
void flip4(unsigned int *src, unsigned int *dest, int n) {
  int i;

  for (i=0;i<n;i++) {
     register unsigned int tmp = src[i];
     dest[i] = FLIP4(tmp);
  }
}


#ifdef cray

/*
 * THESE ROUTINES MUST BE COMPILED ON THE CRAY ONLY SINCE THEY
 * REQUIRE 8-BYTES PER C-TYPE LONG                             
 */

/* Cray to IEEE single precision */
static void c_to_if(t, f) 
long *t,*f;
{
    if (*f != 0) {
        *t = (((*f & 0x8000000000000000) |      /* sign bit */
                 ((((*f & 0x7fff000000000000) >> 48)-16258) << 55)) + /* exp */
                 (((*f & 0x00007fffff000000) +
                    ((*f & 0x0000000000800000) << 1)) << 8));  /* mantissa */
    }
    else *t = *f;
}

#define C_TO_IF(T, F)							\
	if (F != 0) {							\
		T = (((F & 0x8000000000000000) |			\
		((((F & 0x7fff000000000000) >> 48)-16258) << 55)) +	\
		(((F & 0x00007fffff000000) +				\
		((F & 0x0000000000800000) << 1)) << 8));		\
	} else {							\
		T = F;							\
	}

/* IEEE single precison to Cray */
static void if_to_c(t, f)
long *t,*f;
{
    if (*f != 0) {
        *t = (((*f & 0x8000000000000000) |
                ((*f & 0x7f80000000000000) >> 7) +
                (16258 << 48)) |
                (((*f & 0x007fffff00000000) >> 8) | (0x0000800000000000)));
        if ((*f << 1) == 0) *t = 0;
    }
    else *t = *f;
}

/* T and F must be longs! */
#define IF_TO_C( T, F )							\
	if (F != 0) {							\
		T = (((F & 0x8000000000000000) |			\
		((F & 0x7f80000000000000) >> 7) +			\
		(16258 << 48)) |					\
		(((F & 0x007fffff00000000) >> 8) | (0x0000800000000000)));  \
		if ((F << 1) == 0) T = 0;				\
	} else {							\
		T = F;							\
	}

#endif /* cray */



/**************************************************************
 *         I/O Functions    
 **************************************************************/

/*
 * Write a 4-byte integer.
 * Input:  f - the file descriptor
 *         i - the integer
 * Return: 1 = ok, 0 = error
 */
int write_int4(int f, int i) {

#ifdef cray
   i = i << 32;
   return write(f, &i, 4) > 0;
#else
#  ifdef LITTLE
     i = FLIP4(i);
#  endif
   return write(f, &i, 4) > 0;
#endif
}

/*
 * Write a 4-byte IEEE float.
 * Input:  f - the file descriptor
 *         x - the float
 * Return:  1 = ok, 0 = error
 */
int write_float4(int f, float x) {

#ifdef cray
   char buffer[8];
   c_to_if(buffer, &x);
   return write(f, buffer, 4) > 0;
#else
#  ifdef LITTLE
      float y;
      unsigned int *iptr = (unsigned int *) &y, temp;
      y = (float) x;
      temp = FLIP4(*iptr);
      return write(f, &temp, 4) > 0;
#  else
      float y;
      y = (float) x;
      return write(f, &y, 4) > 0;
#  endif
#endif
}

/*
 * receive_data 
 *  Receive a block of data by repeatedly reading from a socket.  This
 *  function will block the process if data is not available.
 *  Input:  socket - socket to read from
 *          buffer - pointer to data buffer
 *          bytes - how many bytes to read into buffer
 *  Return:  number of bytes received.  This will always either be
 *           the same as the bytes requested or 0 if the connection
 *           was lost.
 */

int receive_data(f, buffer, bytes) 
int f;
char *buffer;
int bytes;
{
  int sofar, remaining, len;

  sofar = 0;
  remaining = bytes;
  do {
     len = read( f, buffer + sofar, remaining);
     if (len<=0) return 0;
     sofar += len;
     remaining -= len;
  } while (remaining > 0);
  return bytes;
}

/*
 * Read a 4-byte integer. This function will
 * block if the data is not available.
 * Input:  f - the file descriptor to read from
 *         i - pointer to integer to put result into.
 * Return:  1 = ok, 0 = error
 */
int read_int4(int f, int *i) {

#ifdef LITTLE
   /* read big endian and convert to little endian */
   unsigned int n;
   if (receive_data(f, &n, 4) == 4) {
      *i = FLIP4(n);
      return 1;
   } else {
      return 0;
   }
#else
   if (receive_data(f, i, 4) == 4) {
#  ifdef cray
      *i = *i >> 32;
#  endif
      return 1;
   } else {
      return 0;
   }
#endif
}


/*
 * Read a 4-byte IEEE float.
 * Input:  f - the file descriptor to read from.
 *         x - pointer to float to put result into.
 * Return:  1 = ok, 0 = error
 */
int read_float4(int f, float *x) {

#ifdef cray
   long buffer = 0;

   if (receive_data(f, &buffer, 4) == 4) {
      /* convert IEEE float (buffer) to Cray float (x) */
      if_to_c(x, &buffer);
      return 1;
    }
    return 0;
#else
#  ifdef LITTLE
      unsigned int n, *iptr;
      if (receive_data(f, &n, 4) == 4) {
	 iptr = (unsigned int *) x;
	 *iptr = FLIP4(n);
	 return 1;
      } else {
	 return 0;
      }
#  else
      if (receive_data(f, x, 4) == 4) {
	 return 1;
      } else {
	 return 0;
      }
#  endif
#endif
}

/*
 * initialize the socket
 * return socket descriptor. If failed return -1
 */
int initialize() {
   int sockfd;

   char *hostname;
   char *rawport;
   int port;

   struct  hostent *h;
   struct sockaddr_in serv_addr;

   /** 
    * first read system environment JSERHOST and JSERPORT
    */
   hostname = getenv("JSERHOST");
   rawport = getenv("JSERPORT");

   if (hostname == NULL) {
      printf("Unknown server host name error ...\n");
      return -1;
   }
   
   if (rawport == NULL) 
      port = 7653;
   else
      port = atoi(rawport);

   if ((h=gethostbyname(hostname)) == NULL) {  /* get the host info */
      fprintf(stderr,"initialize: gethostbyname(%s) failed.\n",hostname);
      return -1;
   }

   /*
    * fill in the structure "serv_addr" with the address 
    * of the server that we want to connect with 
    */
   bzero((char *) &serv_addr, sizeof(serv_addr));
   serv_addr.sin_family = AF_INET;
   serv_addr.sin_addr.s_addr = 
                    inet_addr(inet_ntoa(*((struct in_addr *)h->h_addr)));
   serv_addr.sin_port = htons(port);

   /*
   printf("Host name  : %s\n", h->h_name);
   printf("IP Address : %s\n",inet_ntoa(*((struct in_addr *)h->h_addr)));
   */

   /*
    * open a TCP socket (an Internet stream socket).
    */
   if ((sockfd = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
      printf("Can't open stream socket.\n");
      return -1;
   }

   /*
    * connect to server
    */
   if (connect(sockfd, (struct sockaddr *) &serv_addr, 
                                    sizeof(serv_addr)) < 0) {
      printf("Can't connect to sv2 server.\n");
      return -1;
   }

   return sockfd;
}


/*************************************************************
 * sv2_ser_
 * This program writes the x and fx values to socket.
 *
 * input parameters
 *   name - name of the file, it will be used fore title
 *   time - time value
 *   x    - x value array
 *   fx   - fx value array
 *   nx   - size of x and fx array
 ************************************************************/
int sv2_ser_(name, time, x, fx, nx) 
  char *name;
  double *time;
  double x[];
  double fx[];
  int *nx;
{
   int sockfd;
   int size = *nx, name_length, i;
   double ft = *time;
   float ftime = (float)ft;
   int command = RAW_DATA_FEED;
   char ok = 1;
   char none = 0;

   sockfd = initialize();
   if (sockfd == -1)
     return 1;

   /* first send command */
   write_int4(sockfd, command);

   /* send title information */
   name_length = strlen(name);
   write_int4(sockfd, name_length);
   write(sockfd, name, name_length);

   /* write the number of time step and data type */
   write_int4(sockfd, 1);
   write_int4(sockfd, POINT);

   /* write time and nx */
   write_float4(sockfd, ftime);
   write_int4(sockfd, size);

   /* this is not bb */
   write(sockfd, &none, 1);

   /* write x and fx value */
   for (i=0; i<size; i++) {
      write_float4(sockfd, (float)x[i]);
      write_float4(sockfd, (float)fx[i]);
   }

   /* dummys */
   write(sockfd, &none, 1);
   write(sockfd, &none, 1);
   write_int4(sockfd, 0);
   
   close(sockfd);
   return 0;
}



/*************************************************************
 * sv2_sercon_
 * This program writes the connection lists to socket.
 *
 * input parameters
 *   name - name of the file, it will be used fore title
 *   time - time value
 *   x    - x value array
 *   y    - y value array
 *   c    - color value array
 *   nx   - size of x, y, c array
 *   cl   - connection list
 *   nc   - the number of elements of connection list
 ************************************************************/
int sv2_sercon_(name, time, x, y, c, nx, cl, nc) 
  char *name;
  double *time;
  double x[];
  double y[];
  double c[];
  int *nx;
  int cl[];
  int *nc;
{
   int sockfd;
   int size = *nx, con = *nc, name_length, i;
   double ft = *time;
   float ftime = (float)ft;
   int command = RAW_DATA_FEED;
   char ok = 1;
   char none = 0;

   sockfd = initialize();
   if (sockfd == -1)
     return 1;

   /* first send command */
   write_int4(sockfd, command);

   /* send title information */
   name_length = strlen(name);
   write_int4(sockfd, name_length);
   write(sockfd, name, name_length);

   /* write the number of time step and data type */
   write_int4(sockfd, 1);
   write_int4(sockfd, CONNECTIVITY);

   /* write time and nx, nc */
   write_float4(sockfd, ftime);
   write_int4(sockfd, size);

   /* this is not bb */
   write(sockfd, &none, 1);

   /* write x and fx value */
   for (i=0; i<size; i++) {
      write_float4(sockfd, (float)x[i]);
      write_float4(sockfd, (float)y[i]);
   }

   /* write color array */
   write(sockfd, &ok, 1);
   for (i=0; i<size; i++) 
      write_float4(sockfd, (float)c[i]);

   write_int4(sockfd, con);
   for (i=0; i<con; i++) 
      write_int4(sockfd, cl[i]);

   write_int4(sockfd, 0);
   
   close(sockfd);
   return 0;
}


/*************************************************************
 * sv2_mser_
 * This program writes whole set of time data with the x and fx values
 * to socket.
 *
 * input parameters
 *   name - name of the file, it will be used for title
 *   nt   - number of time
 *   time - time value array
 *   x    - x value array
 *   fx   - fx value array
 *   nx   - size of x and fx array
 ************************************************************/
int sv2_mser_(name, nt, time, x, fx, nx) 
  char *name;
  int *nt;
  double time[];
  double x[];
  double fx[];
  int nx[];
{
   int sockfd;
   int ntsize = *nt;
   int name_length, i, j, index;
   int command = RAW_DATA_FEED;
   char ok = 1;
   char none = 0;

   sockfd = initialize();
   if (sockfd == -1)
     return 1;

   /* first send command */
   write_int4(sockfd, command);
   
   /* send title information */
   name_length = strlen(name);
   write_int4(sockfd, name_length);
   write(sockfd, name, name_length);

   /* send nt */
   write_int4(sockfd, ntsize);

   /* write time data */
   index = 0;

   for (i=0; i<ntsize; i++) {
      write_int4(sockfd, POINT);
      write_float4(sockfd, (float)time[i]);
      write_int4(sockfd, nx[i]);

      /* this is not bb */
      write(sockfd, &none, 1);

      for (j=0; j<nx[i]; j++) {
         write_float4(sockfd, (float)x[index]);
         write_float4(sockfd, (float)fx[index]);
         index++;
      }

      /* dummys */
      write(sockfd, &none, 1);
      write(sockfd, &none, 1);
      write_int4(sockfd, 0);
   }
   
   close(sockfd);
   return 0;
}

/*************************************************************
 * sv2_bbser_
 * This program writes the x, y and fxy values to socket.
 *
 * input parameters
 *   name - name of the file, it will be used for title
 *   time - time value
 *   bb    - x min and max  value
 *   xmax    - x max value
 *   nx      - number of x points 
 *   fx     - fxy value array
 ************************************************************/
int sv2_bbser_(name, time, bb, nx, fx) 
  char *name;
  double *time;
  double bb[];
  int *nx;
  double fx[];
{
   int sockfd;
   int size = *nx, name_length, i;
   double ft = *time;
   float ftime = (float)ft;
   int command = RAW_DATA_FEED;
   char ok = 1;
   char none = 0;

   sockfd = initialize();
   if (sockfd == -1)
     return 1;

   /* first send command */
   write_int4(sockfd, command);
   
   /* send title information */
   name_length = strlen(name);
   write_int4(sockfd, name_length);
   write(sockfd, name, name_length);

   /* write the number of time step and data type */
   write_int4(sockfd, 1);
   write_int4(sockfd, POINT);

   /* write time and nx */
   write_float4(sockfd, ftime);
   write_int4(sockfd, size);

   /* this is a bb */
   write(sockfd, &ok, 1);

   write_float4(sockfd, (float)bb[0]);
   write_float4(sockfd, (float)bb[1]);

   /* write fx value */
   for (i=0; i<size; i++) {
      write_float4(sockfd, (float)fx[i]);
   }

   /* dummys */
   write(sockfd, &none, 1);
   write(sockfd, &none, 1);
   write_int4(sockfd, 0);
     
   close(sockfd);
   return 0;
}

/*************************************************************
 * sv2_bbmser_
 * This program writes whole set of time data with the x, y and fxy values
 * to socket.
 *
 * input parameters
 *   name - name of the file, it will be used fore title
 *   nt   - number of time
 *   time - time value array
 *   bb    - x min and max array
 *   nx    -   nx value array
 *   fx   - fxy value array
 ************************************************************/
int sv2_bbmser_(name, nt, time, bb, nx, fx) 
  char *name;
  int *nt;
  double time[];
  double bb[][2];
  int nx[];
  double fx[];
{
   int sockfd;
   int ntsize = *nt, name_length, i, j, index;
   int command = RAW_DATA_FEED;
   char ok = 1;
   char none = 0;


   sockfd = initialize();
   if (sockfd == -1)
     return 1;

   /* first send command */
   write_int4(sockfd, command);
   
   /* send title information */
   name_length = strlen(name);
   write_int4(sockfd, name_length);
   write(sockfd, name, name_length);

   /* send nt */
   write_int4(sockfd, ntsize);

   index = 0;
   for (i=0; i<ntsize; i++) {
      write_int4(sockfd, POINT);
      write_float4(sockfd, (float)time[i]);
      write_int4(sockfd, nx[i]);

      /* this is a bb */
      write(sockfd, &ok, 1);
      write_float4(sockfd, (float)bb[i][0]);
      write_float4(sockfd, (float)bb[i][1]);
     
      for (j=0; j<nx[i]; j++) {
         write_float4(sockfd, (float)fx[index]);
         index++;
      }

      /* dummys */
      write(sockfd, &none, 1);
      write(sockfd, &none, 1);
      write_int4(sockfd, 0);
   }
   
   close(sockfd);
   return 0;
}

/*************************************************************
 * sv2_ser2d_
 * This program writes the x, y and fxy values to socket.
 *
 * input parameters
 *   name - name of the file, it will be used fore title
 *   time - time value
 *   x    - x value array
 *   y    - y value array
 *   fxy  - fxy value array
 *   nx   - size of x and fx array
 ************************************************************/
int sv2_ser2d_(name, time, x, y, fxy, nx) 
  char *name;
  double *time;
  double x[];
  double y[];
  double fxy[];
  int *nx;
{
   printf("Not support yet ...\n");
   return 1;
}


/*************************************************************
 * sv2_mser2d_
 * This program writes whole set of time data with the x, y and fxy values
 * to socket.
 *
 * input parameters
 *   name - name of the file, it will be used fore title
 *   nt   - number of time
 *   time - time value array
 *   x    - x value array
 *   y    - y value array
 *   fxy   - fxy value array
 *   nx   - size of x and fx array
 ************************************************************/
int sv2_mser2d_(name, nt, time, x, y, fxy, nx) 
  char *name;
  int *nt;
  double time[];
  double x[];
  double y[];
  double fxy[];
  int nx[];
{
   printf("Not support yet ...\n");
   return 1;
}


/*************************************************************
 * sv2_bbser2d_
 * sends 2d bounding box and fxy values to socket.
 *
 * input parameters
 *   name - name of the file, it will be used fore title
 *   time - time value
 *   bb    - x min, max, y min, and y max
 *   xmax    - x max value
 *   nx      - number of x points 
 *   ymin    - y min value
 *   ymax    - y max value
 *   ny      - number of y points 
 *   fxy     - fxy value array
 ************************************************************/
int sv2_bbser2d_(name, time, bb, nx, ny, fxy) 
  char *name;
  double *time;
  double bb[];
  int *nx;
  int *ny;
  double fxy[];
{
   int sockfd;
   int size, xsize = *nx, ysize = *ny, name_length, i;
   double ft = *time;
   float ftime = (float)ft;
   int command = RAW_DATA_FEED;
   char ok = 1;
   char none = 0;

   sockfd = initialize();
   if (sockfd == -1)
     return 1;

   /* first send command */
   write_int4(sockfd, command);
   
   /* send title information */
   name_length = strlen(name);
   write_int4(sockfd, name_length);
   write(sockfd, name, name_length);

   /* write the number of time step and data type */
   write_int4(sockfd, 1);
   write_int4(sockfd, TWOD);

   size = xsize * ysize;
   /* write time, x and y size, and bb */
   write_float4(sockfd, ftime);
   write_int4(sockfd, xsize);
   write_int4(sockfd, ysize);
   write_int4(sockfd, size);

   /* this is a bb */
   write(sockfd, &ok, 1);

   write_float4(sockfd, (float)bb[0]);
   write_float4(sockfd, (float)bb[1]);
   write_float4(sockfd, (float)bb[2]);
   write_float4(sockfd, (float)bb[3]);

   /* write fx value */
   for (i=0; i<size; i++) {
      write_float4(sockfd, (float)fxy[i]);
   }

   /* dummys */
   write(sockfd, &none, 1);
   write_int4(sockfd, 0);

   close(sockfd);
   return 0;
}

/*************************************************************
 * sv2_bbmser2d_
 * c program to link with fortran 90
 * This program writes whole set of time data with the x, y and fxy values
 * to socket.
 *
 * input parameters
 *   name - name of the file, it will be used fore title
 *   nt   - number of time
 *   time - time value array
 *   bb    - x min, max, y min, and y max value array
 *   nx    -   nx value array
 *   ny    -   ny value array
 *   fxy   - fxy value array
 ************************************************************/
int sv2_bbmser2d_(name, nt, time, bb, nx, ny, fxy) 
  char *name;
  int *nt;
  double time[];
  double bb[][4];
  int nx[];
  int ny[];
  double fxy[];

{
   int sockfd;
   int ntsize = *nt, size, name_length, i, j, index;
   int command = RAW_DATA_FEED;
   char ok = 1;
   char none = 0;


   sockfd = initialize();
   if (sockfd == -1)
     return 1;

   /* first send command */
   write_int4(sockfd, command);
   
   /* send title information */
   name_length = strlen(name);
   write_int4(sockfd, name_length);
   write(sockfd, name, name_length);

   /* send nt */
   write_int4(sockfd, ntsize);

   index = 0;
   for (i=0; i<ntsize; i++) {      
      write_int4(sockfd, TWOD);
      write_float4(sockfd, time[i]);
      write_int4(sockfd, nx[i]);
      write_int4(sockfd, ny[i]);
      size = nx[i] * ny[i];
      write_int4(sockfd, size);

      /* this is a bb */
      write(sockfd, &ok, 1);
      write_float4(sockfd, (float)bb[i][0]);
      write_float4(sockfd, (float)bb[i][1]);
      write_float4(sockfd, (float)bb[i][2]);
      write_float4(sockfd, (float)bb[i][3]);
     
      for (j=0; j<size; j++) {
         write_float4(sockfd, (float)fxy[index]);
         index++;
      }
 
      /* dummys */
      write(sockfd, &none, 1);
      write_int4(sockfd, 0);
   }
   
   close(sockfd);
   return 0;
}


/*************************************************************
 * sv2_ser3d_
 * This program writes the xsize, ysize, zsize and z values to socket.
 *
 * input parameters
 *   name - name of the file, it will be used for title
 *   time - time value
 *   nx    - x dimension 
 *   ny    - y dimension 
 *   nz    - z dimension 
 *   z   - z array
 ************************************************************/
int sv2_ser3d_(name, time, nx, ny, nz, grid) 
  char *name;
  double *time;
  int *nx;
  int *ny;
  int *nz; 
  double grid[];
{
   int sockfd;
   int nnx = *nx, nny = *ny, nnz = *nz, name_length, i;
   int size = nnx * nny * nnz;
   double ft = *time;
   float ftime = (float)ft;   
   int command = RAW_DATA_FEED;

   sockfd = initialize();
   if (sockfd == -1)
     return 1;

   /* first send command */
   write_int4(sockfd, command);

   /* send title information */
   name_length = strlen(name);
   write_int4(sockfd, name_length);
   write(sockfd, name, name_length);

   /* write the number of time step and data type */
   write_int4(sockfd, 1);
   write_int4(sockfd, THREED);

   /* write time and sizes */
   write_float4(sockfd, ftime);
   write_int4(sockfd, nnx);
   write_int4(sockfd, nny);
   write_int4(sockfd, nnz);
   write_int4(sockfd, size);

   /* write x and fx value */
   for (i=0; i<size; i++) 
      write_float4(sockfd, (float)grid[i]);
   
   close(sockfd);
   return 0;
}


/*************************************************************
 * sv2_mser3d_
 * This program writes the xsize, ysize, zsize and z values to socket.
 *
 * input parameters
 *   name - name of the file, it will be used for title
 *   time - time value
 *   nx    - x dimension 
 *   ny    - y dimension 
 *   nz    - z dimension 
 *   z   - z array
 ************************************************************/

int sv2_mser3d_(name, nt, time, nx, ny, nz, grid) 
  char *name;
  int *nt;
  double time[];
  int nx[];
  int ny[];
  int nz[]; 
  double grid[];
{
   int sockfd;
   int ntsize = *nt;
   int name_length, index, i, j, k, size;
   int command = RAW_DATA_FEED;

   sockfd = initialize();
   if (sockfd == -1)
     return 1;

   /* first send command */
   write_int4(sockfd, command);

   /* send title information */
   name_length = strlen(name);
   write_int4(sockfd, name_length);
   write(sockfd, name, name_length);

   /* send no of time */
   write_int4(sockfd, ntsize);    

   /* write time data */
   index = 0;
   for (i=0; i<ntsize; i++) {
      write_int4(sockfd, THREED);

      write_float4(sockfd, (float)time[i]);
      write_int4(sockfd, nx[i]);
      write_int4(sockfd, ny[i]);
      write_int4(sockfd, nz[i]);
      size = nx[i] * ny[i] * nz[i];
      write_int4(sockfd, size);

      /* grid value */
      for (j=0; j<size; j++) {
         write_float4(sockfd, (float)grid[index]);
         index++;
      }   
   }
   
   close(sockfd);
   return 0;
}


int sv2_ser(char *name, double time, double x[], double y[], int n) {
   return sv2_ser_(name, &time, x, y, &n);
}

int sv2_sercon(char *name, double time, double x[], double y[], double c[], int nx, int cl[], int nc) {
   sv2_sercon_(name, &time, x, y, c, &nx, cl, &nc);
}

int sv2_mser(char *name, int nt, double time[], double x[], double fx[], int nx[]) {
   return sv2_mser_(name, &nt, time, x, fx, nx);
}

int sv2_bbser(char *name, double time, double bb[], int nx, double fx[]) {
  sv2_bbser_(name, &time, bb, &nx, fx);
}

int sv2_bbmser(char *name, int nt, double time[], double bb[][2], int nx[], double fx[]) {
   sv2_bbmser_(name, &nt, time, bb, nx, fx);
}

int sv2_ser2d(char *name, double time, double x[], double y[], double fxy[], int n) {
   return sv2_ser2d_(name, &time, x, y, fxy, &n);
}

int sv2_mser2d(char *name, int nt, double time[], double x[], double y[], double fxy[], int nx[]) {
   return sv2_mser2d_(name, &nt, time, x, y, fxy, nx);
}

int sv2_bbser2d(char *name, double time, double bb[], int nx, int ny, double fxy[]) {
  sv2_bbser2d_(name, &time, bb, &nx, &ny, fxy);
}

int sv2_bbmser2d(char *name, int nt, double time[], double bb[][4], int nx[], int ny[], double fxy[]) {
   sv2_bbmser2d_(name, &nt, time, bb, nx, ny, fxy);
}

int sv2_ser3d(char *name, double time, int nx, int ny, int nz, double grid[]) {
   return sv2_ser3d_(name, &time, &nx, &ny, &nz, grid);
}

int sv2_mser3d(char *name, int nt, double time[], int nx[], int ny[], int nz[], double grid[]) {
   return sv2_mser3d_(name, &nt, time, nx, ny, nz, grid);
}

