! This routine  will do a 3 dimensional interpolation to 3rd order using a 3 dimensional
! Taylor series expansion
! nx,ny,nz= number of points in i,j,k directions
! x(nx),y(ny),z(nz) = x,y,z arrays
! n = number of interpolation points
! intlist[n][3] = x,y,z points in a list, these are the interpolation points
! storage = output = LUD decomposition array with i,j,k values stored, the array looks like
!
!  | f0 |      =      | 1  a0  b0  c0  a0^2/2  b0^2/2  c0^2/2   a0b0  a0c0  b0c0  |        | f   |
!  | f1 |      =      | 1  a1  b1  c1  a1^2/2  b1^2/2  c1^2/2   a1b1  a1c1  b1c1  |        | fx  |
!  | f2 |      =      | 1  a2  b2  c2  a2^2/2  b2^2/2  c2^2/2   a2b2  a2c2  b2c2  |        | fy  |
!  | f3 |      =      | 1  a3  b3  c3  a3^2/2  b3^2/2  c3^2/2   a3b3  a3c3  b3c3  |        | fz  |
!  | f4 |      =      | 1  a4  b4  c4  a4^2/2  b4^2/2  c4^2/2   a4b4  a4c4  b4c4  |        | fxx |
!  | f5 |      =      | 1  a5  b5  c5  a5^2/2  b5^2/2  c5^2/2   a5b5  a5c5  b5c5  |        | fyy |
!  | f6 |      =      | 1  a6  b6  c6  a6^2/2  b6^2/2  c6^2/2   a6b6  a6c6  b6c6  |        | fzz |
!  | f7 |      =      | 1  a7  b7  c7  a7^2/2  b7^2/2  c7^2/2   a7b7  a7c7  b7c7  |        | fxy |
!  | f8 |      =      | 1  a8  b8  c8  a8^2/2  b8^2/2  c8^2/2   a8b8  a8c8  b8c8  |        | fxz |
!  | f9 |      =      | 1  a9  b9  c9  a9^2/2  b9^2/2  c9^2/2   a9b9  a9c9  b9c9  |        | fyz |

! in the storage will be the LUD decomposition of fhis, along with the extra columns of
! i,j,k,pivot

subroutine int3do3(nx,ny,nz,x,y,z,n,intlist,mask,storage)
  implicit none
  interface

   function get_closet_point(n,x,l)
     implicit none
     integer get_closet_point
     integer :: n
     real*8,dimension(n) :: x
     real*8  :: l
   end function get_closet_point

   function check_point_stencil_diameter(i,j,k,mask,stencil_size)
     implicit none
     integer check_point_stencil_diameter
     integer :: i,j,k
     real*8 :: mask(:,:,:)
     integer :: stencil_size
   end function 

   function is_point_valid(rule_loop,rules,ijkp,ijk,xyzp)
     implicit none
     integer is_point_valid
     integer :: rule_loop
     integer,dimension(3) :: ijkp,ijk,rules,xyzp
   end function

   function reshift_point(nx,ny,nz,x,y,z,list,i,j,k,mask)
     implicit none
     integer reshift_point
     integer :: nx,ny,nz
     real*8,dimension(nx) :: x
     real*8,dimension(ny) :: y
     real*8,dimension(nz) :: z
     real*8,dimension(:) :: list
     integer :: i,j,k
     real*8,dimension(nx,ny,nz) :: mask
   end function reshift_point
   
   subroutine get_points(nx,ny,nz,i,j,k,x,y,z,list,mask,max_cirm,ijksave,radius,points)
     implicit none
     integer :: nx,ny,nz
     integer :: i,j,k
     real*8,dimension(nx) :: x
     real*8,dimension(ny):: y
     real*8,dimension(nz):: z
     real*8,dimension(:) :: list
     real*8,dimension(nx,ny,nz)  :: mask
     integer :: max_cirm
     integer,dimension(:,:) :: ijksave
     real*8,dimension(:) :: radius
     integer :: points
   end subroutine get_points

   subroutine sort_points(n,r,ijksave)
     implicit none
     integer :: n
     real*8,dimension(:)  :: r
     integer,dimension(:,:) :: ijksave
   end subroutine sort_points

   subroutine fill_matrix(row,x,y,z,list,int_points,local_matrix)
     implicit none
     integer :: row
     real*8 :: x,y,z
     real*8,dimension(:) :: list
     integer :: int_points
     real*8 :: local_matrix(:,:)
   end subroutine fill_matrix

  end interface

  integer,intent(IN)  :: nx,ny,nz
  real*8, dimension(nx),intent(IN)  :: x
  real*8, dimension(ny),intent(IN) :: y
  real*8, dimension(nz),intent(IN) :: z
  integer,intent(IN) :: n
  real*8,dimension(n,3),intent(IN) :: intlist
  real*8,dimension(nx,ny,nz),intent(IN) :: mask
  integer,parameter :: stencil_size = 3
  integer,parameter :: int_points = 10
  real*8,dimension(int_points,int_points+4,n),intent(INOUT) :: storage
  
  !==============================================================================
  ! local storage here
  real*8,dimension(int_points,int_points) :: local_matrix
  integer,dimension(int_points) :: pivot
  integer,dimension(int_points,3) :: local_ijk
  integer,dimension(int_points,3) :: rules
  real*8,dimension(3) :: list
  data rules / 0,1,0,0,2,0,0,1,1,0,  0,0,1,0,0,2,0,1,0,1,  0,0,0,1,0,0,2,0,1,1 /
  real*8 :: a,b,c
  
  integer :: point_loop,rule_loop
  integer :: points
  logical :: done_int_loop
  integer :: max_cirm_search
  integer,parameter :: mcss = 9 ! I set this to be 9 for a greater search range
  integer,parameter :: point_store_size= mcss*mcss*mcss!  need 5^5 possible points
  integer,parameter :: max_cirm = 3
  integer,dimension(point_store_size,3) :: ijksave,ijkp
  real*8, dimension(point_store_size) :: radius
  integer,dimension(point_store_size) :: sort_index
  integer,dimension(3) :: xyzp
  integer :: ierr
  real*8 :: xint,yint,zint
  integer :: lt,shift
  integer :: i,j,k,ierror,int_loop
  !==============================================================================
  
  do point_loop = 1, n                 !loop over all of the interpolation points
     list(:) = intlist(point_loop,:)
     i = get_closet_point(nx,x,list(1))
     j = get_closet_point(ny,y,list(2))
     k = get_closet_point(nz,z,list(3))
     ! if the i,j,k has a mask of 0, move it 
     shift = reshift_point(nx,ny,nz,x,y,z,list,i,j,k,mask)
     if (shift.gt.0) write(6,*) point_loop,shift
     max_cirm_search = max_cirm + shift
     ierror = check_point_stencil_diameter(i,j,k,mask,stencil_size)
     ! ignore this test for now, and just print an error if there is a problem
     if (ierror.ne.0) write(*,*) 'Error:: interpolation for point',point_loop
     call get_points(nx,ny,nz,i,j,k,x,y,z,list,mask,max_cirm_search,ijksave,radius,points)
     call sort_points(points,radius,ijksave)
     do lt = 1 , points
        ijkp(lt,:) = ijksave(lt,:) - ijksave(1,:) ! This is how far off the stencil point is
     end do
     do rule_loop = 1,int_points
        int_loop = 0; done_int_loop = .true.
        if (rule_loop > int_points - 3) then ! see if it's one-sided in either direction
           xyzp = 0
           if ( (local_ijk(2,1) > local_ijk(1,1)).and. &
                (local_ijk(5,1) > local_ijk(1,1))) xyzp(1) = 1
           if ( (local_ijk(2,1) < local_ijk(1,1)).and. &
                (local_ijk(5,1) < local_ijk(1,1))) xyzp(1) = -1

           if ( (local_ijk(3,2) > local_ijk(1,2)).and. &
                (local_ijk(6,2) > local_ijk(1,2))) xyzp(2) = 1
           if ( (local_ijk(3,2) < local_ijk(1,2)).and. &
                (local_ijk(6,2) < local_ijk(1,2))) xyzp(2) = -1

           if ( (local_ijk(4,3) > local_ijk(1,3)).and. &
                (local_ijk(7,3) > local_ijk(1,3))) xyzp(3) = 1
           if ( (local_ijk(4,3) < local_ijk(1,3)).and. &
                (local_ijk(7,3) < local_ijk(1,3))) xyzp(3) = -1

        end if
        
        do while (done_int_loop)
           int_loop = int_loop + 1
           if (int_loop > points) then
              write(6,*) 'int3do3 failed at',point_loop,rule_loop
              stop
           end if
           if (rule_loop == 1) then 
              local_ijk(1,:) = ijksave(1,:)
              
              call fill_matrix(1,x(ijksave(1,1)),y(ijksave(1,2)),z(ijksave(1,3)),&
                   list,int_points,local_matrix)
              done_int_loop = .false.
              ijksave(1,:) = 0
           else if (ijksave(int_loop,1).ne.0) then
              ierr = is_point_valid(rule_loop,rules(rule_loop,:),ijkp(int_loop,:),&
                   ijksave(int_loop,:),xyzp)
              if (ierr == 0) then                         !this is a good point
                 local_ijk(rule_loop,:) = ijksave(int_loop,:)
                 call fill_matrix(rule_loop,x(ijksave(int_loop,1)),           &
                      y(ijksave(int_loop,2)),z(ijksave(int_loop,3)),          &
                      list,int_points,local_matrix)
                 done_int_loop = .false.
                 ijksave(int_loop,:) = 0
              end if
           end if
        end do ! for the while loop
     end do ! rule_loop
     ! now we have local_matrix, and local_ijk, so now we do an LUD decomposition
!!$     write(6,*) '-------------------------',point_loop
     call dgefa(local_matrix,int_points,int_points,pivot,ierr) 
     if (ierr.ne.0) then
        write(*,*) 'Error in int3do3', point_loop
        stop
     end if
     storage(1:int_points,1:int_points,point_loop) = local_matrix(:,:)
     storage(1:int_points,int_points+1,point_loop) = local_ijk(:,1)
     storage(1:int_points,int_points+2,point_loop) = local_ijk(:,2)
     storage(1:int_points,int_points+3,point_loop) = local_ijk(:,3)
     storage(1:int_points,int_points+4,point_loop) = pivot(:)
   end do !point_loop
   end subroutine

   function get_closet_point(n,x,l)
     integer get_closet_point
     integer :: n
     real*8,dimension(n) :: x
     real*8  :: l
     
     integer :: i
     real*8  :: mindiff,diff
     mindiff = abs(l - x(1))
     get_closet_point = 1
     do i = 1, n
        diff = abs(l - x(i))
        if ( diff<mindiff) then
           get_closet_point = i
           mindiff = diff
        end if
     end do
   end function get_closet_point
   
   function reshift_point(nx,ny,nz,x,y,z,list,i,j,k,mask)
     implicit none
     interface
        
        subroutine get_points(nx,ny,nz,i,j,k,x,y,z,list,mask,max_cirm,ijksave,radius,points)
          implicit none
          integer :: nx,ny,nz
          integer :: i,j,k
          real*8,dimension(nx) :: x
          real*8,dimension(ny):: y
          real*8,dimension(nz):: z
          real*8,dimension(:) :: list
          real*8  :: mask(nx,ny,nz)
          integer :: max_cirm
          integer,dimension(:,:) :: ijksave
          real*8,dimension(:) :: radius
          integer :: points
        end subroutine get_points
        
        subroutine sort_points(n,r,ijksave)
          implicit none
          integer :: n
          real*8,dimension(:)  :: r
          integer,dimension(:,:) :: ijksave
        end subroutine sort_points
     end interface
     integer reshift_point
     integer :: nx,ny,nz
     real*8,dimension(nx) :: x
     real*8,dimension(ny) :: y
     real*8,dimension(nz) :: z

     real*8,dimension(:) :: list
     integer :: i,j,k
     real*8,dimension(nx,ny,nz) :: mask
     
     integer,parameter :: max_shift = 2 ! pick this for now
     integer,parameter :: msp1 = max_shift+1
     integer,dimension(msp1*msp1*msp1,3) :: ijksave_local
     real*8,dimension(msp1*msp1*msp1) :: radius_local
     integer :: points_local
     integer :: ishift,jshift,kshift
     reshift_point = 0
     if (mask(i,j,k) == 0) then ! this is an arbitrary choice in preffered shifting of point
        reshift_point = 1
        call get_points(nx,ny,nz,i,j,k,x,y,z,list,mask,2,ijksave_local,radius_local,points_local)
        call sort_points(points_local,radius_local,ijksave_local)        
        ishift = abs(i -  ijksave_local(1,1))
        jshift = abs(j -  ijksave_local(1,2))
        kshift = abs(k -  ijksave_local(1,3))
        reshift_point = max(ishift,jshift,kshift)
        i = ijksave_local(1,1)
        j = ijksave_local(1,2)
        k = ijksave_local(1,3)
     end if

   end function reshift_point
     
   subroutine get_points(nx,ny,nz,i,j,k,x,y,z,list,mask,max_cirm,ijksave,radius,points)
     implicit none
     integer :: nx,ny,nz
     integer :: i,j,k
     real*8,dimension(nx) :: x
     real*8,dimension(ny):: y
     real*8,dimension(nz):: z
     real*8,dimension(:) :: list
     real*8  :: mask(nx,ny,nz)
     integer :: max_cirm
     integer,dimension(:,:) :: ijksave
     real*8,dimension(:) :: radius
     integer :: points

     integer :: il,jl,kl

     points = 0
     do kl = max(1,k-max_cirm+1),min(nz,k+max_cirm-1)
        do jl = max(1,j-max_cirm+1),min(ny,j+max_cirm-1)
           do il = max(1,i-max_cirm+1),min(nx,i+max_cirm-1)
              if (mask(il,jl,kl).ne.0.0) then
                 points = points + 1
                 ijksave(points,1) = il
                 ijksave(points,2) = jl
                 ijksave(points,3) = kl
                 radius(points) =(x(il)-list(1))**2+(y(jl)-list(2))**2+(z(kl)-list(3))**2
              end if
           end do
        end do
     end do
   end subroutine get_points
   
   subroutine sort_points(n,r,ijksave)
     implicit none
     integer :: n
     integer,dimension(n) :: index
     real*8,dimension(:)  :: r
     integer,dimension(:,:) :: ijksave

     integer,dimension(n,3) :: ijktemp
     real*8,dimension(n) :: rtemp
     integer l

     do l = 1, n
        index(l) = l
     end do
     
     call batcher(r,index,n)
     ijktemp(:,:) = ijksave(:,:)
     rtemp(:) = r
     do l = 1, n
        ijksave(l,:) = ijktemp(index(l),:)
        r(l) = rtemp(index(l))
     end do

   end subroutine sort_points

   subroutine fill_matrix(row,x,y,z,list,int_points,local_matrix)
     implicit none
     integer :: row
     real*8 :: x,y,z
     real*8,dimension(:) :: list
     integer :: int_points
     real*8 :: local_matrix(:,:)
     
     real*8 :: a,b,c

     a = list(1) - x
     b = list(2) - y
     c = list(3) - z

     local_matrix(row,1) = 1.0
     local_matrix(row,2) = a
     local_matrix(row,3) = b
     local_matrix(row,4) = c
     local_matrix(row,5) = a*a*0.5
     local_matrix(row,6) = b*b*0.5
     local_matrix(row,7) = c*c*0.5
     local_matrix(row,8) = a*b
     local_matrix(row,9) = a*c
     local_matrix(row,10) = b*c
!!$     write(6,11)  local_matrix(row,1),local_matrix(row,2),local_matrix(row,3), local_matrix(row,4) , &
!!$          local_matrix(row,5) , local_matrix(row,6) , local_matrix(row,7) , local_matrix(row,8) , &
!!$          local_matrix(row,9) , local_matrix(row,10)
11   format(1P, '[', 10e12.3,']')
   end subroutine fill_matrix
   
   function is_point_valid(rule_loop,rules,ijkp,ijk,xyzp)
     implicit none
     integer is_point_valid
     integer :: rule_loop
     integer,dimension(3) :: ijkp,ijk,rules,xyzp

     ! two tests, one for all terms, the other is for mixed derivatives rule_loop>7
     is_point_valid = 1 ! place non-valid, until we validate it
     if ( (abs(ijkp(1)) .eq.rules(1)).and.(abs(ijkp(2)).eq. rules(2)).and. &
          (abs(ijkp(3)) .eq. rules(3))) then
        if (rule_loop < 8 ) then
           is_point_valid = 0 ! point is valid
        else if (rule_loop == 8) then
              if ( (sign(1,xyzp(1))==sign(1,ijkp(1)).or. xyzp(1)==0).and. &
                 (sign(1,xyzp(2))==sign(1,ijkp(2)) .or.  xyzp(2)==0 )) then
                 is_point_valid = 0
              end if
        else if (rule_loop == 9) then
              if ( (sign(1,xyzp(1)) == sign(1,ijkp(1)).or.xyzp(1)==0).and. &
                 (sign(1,xyzp(3)) == sign(1,ijkp(3)).or.xyzp(3)==0) ) then
                 is_point_valid = 0
              end if
        else if (rule_loop == 10) then
              if ( (sign(1,xyzp(2)) == sign(1,ijkp(2)).or.xyzp(2)==0).and. &
                 (sign(1,xyzp(3)) == sign(1,ijkp(3)).or.xyzp(3)==0) ) then
                 is_point_valid = 0
              end if
           end if
        else if (rule_loop>4.and.rule_loop<8) then
           if ( (2*abs(ijkp(1)) .eq.rules(1)).and.(2*abs(ijkp(2)).eq. rules(2)).and. &
                (2*abs(ijkp(3)) .eq. rules(3))) then
              ! this might be (fxx,fyy, or fzz) and if it is, then the rules for this will
              !be 2, and the valid solution can be just +-,1
              is_point_valid=0
           end if
        end if
   end function is_point_valid

   function check_point_stencil_diameter(i,j,k,mask,stencil_size)
     implicit none
     integer check_point_stencil_diameter
     integer :: i,j,k
     real*8 :: mask(:,:,:)
     integer :: stencil_size
     ! dummy routine which just returns no errors now
     check_point_stencil_diameter = 0
   end function check_point_stencil_diameter

!==============================================================================
! This function will use the storage matrix and the input 3d grid function to
! solve for a 1D array
! nx,ny,nz= number of points in i,j,k directions
! storage = input = LUD decomposition array with i,j,k values stored, the array looks like
! n = number of interpolation points
! uin = 3d grid function, input
! wd = which derivative = (1,2,3,4,5,6,7,8,9,10) =
! f,fx,fy,fz,fxx,fyy,fzz,fxy,fxz,fyz
! NOTE: the order of accuracy is (3,2,2,2,1,1,1,1,1,1)
! uout = 1d grid function (n) , output
subroutine eval3do3(nx,ny,nz,n,storage,wd,uin,uout)
  implicit none
  integer :: nx,ny,nz,n
  integer,parameter :: int_points = 10
  real*8,dimension(int_points,int_points+4,n),intent(INOUT) :: storage
  real*8,dimension(int_points,int_points) :: mat2d
  real*8,dimension(nx,ny,nz) :: uin
  real*8,dimension(n) :: uout
  integer :: wd

  integer,dimension(int_points) :: pivot,i,j,k
  real*8,dimension(int_points) :: rhs
  integer :: l,lin

 
  do l = 1 , n

     mat2d(:,:) = storage(1:int_points,1:int_points,l)
     pivot(:) = storage(1:int_points,int_points+4,l)
     i(:) = storage(1:int_points,int_points+1,l)
     j(:) = storage(1:int_points,int_points+2,l)
     k(:) = storage(1:int_points,int_points+3,l)

     do lin = 1 , int_points
        rhs(lin) = uin(i(lin),j(lin),k(lin))
     end do

     call dgesl(mat2d,int_points,int_points,pivot,rhs,0)

     uout(l) = rhs(wd)

  end do
end subroutine eval3do3
